from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

import pandas as pd
import regex
from lxml import etree

from dsp_tools.commands.excel2xml.excel2xml_lib import append_permissions
from dsp_tools.commands.excel2xml.excel2xml_lib import make_bitstream_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_boolean_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_color_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_date_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_decimal_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_geometry_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_geoname_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_integer_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_link
from dsp_tools.commands.excel2xml.excel2xml_lib import make_list_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_region
from dsp_tools.commands.excel2xml.excel2xml_lib import make_resource
from dsp_tools.commands.excel2xml.excel2xml_lib import make_resptr_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_root
from dsp_tools.commands.excel2xml.excel2xml_lib import make_text_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import make_uri_prop
from dsp_tools.commands.excel2xml.excel2xml_lib import write_xml
from dsp_tools.commands.excel2xml.propertyelement import PropertyElement
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.data_formats.shared import check_notna

# ruff: noqa: E501 (line-too-long)


def _read_cli_input_file(datafile: str) -> pd.DataFrame:
    """
    Parse the input file from the CLI (in either CSV or Excel format)

    Args:
        datafile: path to the input file

    Raises:
        BaseError: if the input file is neither .csv, .xls nor .xlsx

    Returns:
        a pandas DataFrame with the input data
    """
    if regex.search(r"\.csv$", datafile):
        dataframe = pd.read_csv(
            filepath_or_buffer=datafile,
            encoding="utf_8_sig",  # utf_8_sig is the default encoding of Excel
            dtype="str",
            sep=None,
            engine="python",  # let the "python" engine detect the separator
        )
    elif regex.search(r"(\.xls|\.xlsx)$", datafile):
        dataframe = pd.read_excel(io=datafile, dtype="str")
    else:
        raise BaseError(f"Cannot open file '{datafile}': Invalid extension. Allowed extensions: 'csv', 'xls', 'xlsx'")
    return dataframe


def _validate_and_prepare_cli_input_file(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Make sure that the required columns are present,
    replace NA-like cells by NA,
    remove empty columns, so that the max_num_of_props can be calculated without errors,
    and remove empty rows, to prevent them from being processed and raising an error.

    Args:
        dataframe: pandas dataframe with the input data

    Raises:
        BaseError: if one of the required columns is missing

    Returns:
        the prepared pandas DataFrame
    """
    # make sure that the required columns are present
    required_columns = ["id", "label", "restype", "permissions", "prop name", "prop type", "1_value"]
    if any(req not in dataframe for req in required_columns):
        raise BaseError(
            f"Some columns in your input file are missing. The following columns are required: {required_columns}"
        )

    # replace NA-like cells by NA
    dataframe = dataframe.map(
        lambda x: x if pd.notna(x) and regex.search(r"[\p{L}\d_!?\-]", str(x), flags=regex.U) else pd.NA
    )

    # remove empty columns/rows
    dataframe = dataframe.dropna(axis="columns", how="all")
    dataframe = dataframe.dropna(axis="index", how="all")

    return dataframe


def _convert_rows_to_xml(
    dataframe: pd.DataFrame,
    max_num_of_props: int,
) -> list[etree._Element]:
    """
    Iterate through the rows of the CSV/Excel input file,
    convert every row to either a XML resource or an XML property,
    and return a list of XML resources.

    Args:
        dataframe: pandas dataframe with the input data
        max_num_of_props: highest number of properties that a resource in this file has

    Raises:
        BaseError: if one of the rows is neither a resource-row nor a property-row,
            or if the file starts with a property-row

    Returns:
        a list of XML resources (with their respective properties)
    """
    resources: list[etree._Element] = []

    # to start, there is no previous resource
    resource: Optional[etree._Element] = None

    for index, row in dataframe.iterrows():
        row_number = int(str(index)) + 2
        is_resource = check_notna(row["id"])
        is_property = check_notna(row["prop name"])
        match is_resource, is_property:
            case (True, True) | (False, False):  # either the row is a resource-row or a property-row, but not both
                raise BaseError(
                    f"Exactly 1 of the 2 columns 'id' and 'prop name' must be filled. "
                    f"Excel row {row_number} has too many/too less entries:\n"
                    f"id:        '{row['id']}'\n"
                    f"prop name: '{row['prop name']}'"
                )
            case True, False:  # this is a resource-row
                # the previous resource is finished, a new resource begins: append the previous to the resulting list
                # in all cases (except for the very first iteration), a previous resource exists
                if resource is not None:
                    resources.append(resource)
                resource = _convert_resource_row_to_xml(row_number=row_number, row=row)
            case False, True:  # this is a property-row
                if resource is None:
                    raise BaseError(
                        "The first row of your Excel/CSV is invalid. "
                        "The first row must define a resource, not a property."
                    )
                prop = _convert_property_row_to_xml(
                    row_number=row_number,
                    row=row,
                    max_num_of_props=max_num_of_props,
                    resource_id=resource.attrib["id"],
                )
                resource.append(prop)

    # append the resource of the very last iteration of the for loop
    if resource is not None:
        resources.append(resource)

    return resources


def _append_bitstream_to_resource(
    resource: etree._Element,
    row: pd.Series[Any],
    row_number: int,
) -> etree._Element:
    """
    Create a bitstream-prop element, and append it to the resource.
    If the file permissions are missing, try to deduce them from the resource permissions.

    Args:
        resource: the resource element to which the bitstream-prop element should be appended
        row: the row of the CSV/Excel file from where all information comes from
        row_number: row number of the CSV/Excel sheet

    Warning:
        if the file permissions are missing and cannot be deduced from the resource permissions

    Returns:
        the resource element with the appended bitstream-prop element
    """
    file_permissions = row.get("file permissions")
    if not check_notna(file_permissions):
        if resource_permissions := row.get("permissions"):
            file_permissions = resource_permissions
        else:
            file_permissions = ""
            warnings.warn(
                f"Missing file permissions for file "
                f"'{row['file']}' (Resource ID '{row['id']}', Excel row {row_number}). "
                f"An attempt to deduce them from the resource permissions failed."
            )
    resource.append(
        make_bitstream_prop(
            path=str(row["file"]),
            permissions=str(file_permissions),
            calling_resource=row["id"],
        )
    )
    return resource


def _convert_resource_row_to_xml(
    row_number: int,
    row: pd.Series[Any],
) -> etree._Element:
    """
    Convert a resource-row to an XML resource element.
    First, check if the mandatory cells are present.
    Then, call the appropriate function, depending on the restype (Resource, LinkObj, Region).

    Args:
        row_number: row number of the CSV/Excel sheet
        row: the pandas series representing the current row

    Warning:
        if a mandatory cell is missing

    Returns:
        the resource element created from the row
    """
    # read and check the mandatory columns
    resource_id = row["id"]
    resource_label = row.get("label")
    if pd.isna([resource_label]):
        resource_label = ""
        warnings.warn(f"Missing label for resource '{resource_id}' (Excel row {row_number})")
    elif not check_notna(resource_label):
        warnings.warn(
            f"The label of resource '{resource_id}' looks suspicious: '{resource_label}' (Excel row {row_number})"
        )
    resource_restype = row.get("restype")
    if not check_notna(resource_restype):
        resource_restype = ""
        warnings.warn(f"Missing restype for resource '{resource_id}' (Excel row {row_number})")
    resource_permissions = row.get("permissions")
    if not check_notna(resource_permissions):
        resource_permissions = ""
        warnings.warn(f"Missing permissions for resource '{resource_id}' (Excel row {row_number})")

    # construct the kwargs for the method call
    kwargs_resource = {"label": resource_label, "permissions": resource_permissions, "id": resource_id}
    if check_notna(row.get("ark")):
        kwargs_resource["ark"] = row["ark"]
    if check_notna(row.get("iri")):
        kwargs_resource["iri"] = row["iri"]
    if check_notna(row.get("ark")) and check_notna(row.get("iri")):
        warnings.warn(
            f"Both ARK and IRI were provided for resource '{resource_label}' ({resource_id}). "
            "The ARK will override the IRI."
        )
    if check_notna(row.get("created")):
        kwargs_resource["creation_date"] = row["created"]

    # call the appropriate method
    if resource_restype == "Region":
        with warnings.catch_warnings():  # prevent dublette warnings: most problems were already checked above
            warnings.filterwarnings("ignore")
            resource = make_region(**kwargs_resource)
    elif resource_restype == "LinkObj":
        with warnings.catch_warnings():  # prevent dublette warnings: most problems were already checked above
            warnings.filterwarnings("ignore")
            resource = make_link(**kwargs_resource)
    else:
        kwargs_resource["restype"] = resource_restype
        with warnings.catch_warnings():  # prevent dublette warnings: most problems were already checked above
            warnings.filterwarnings("ignore")
            resource = make_resource(**kwargs_resource)
        with warnings.catch_warnings():  # ignore only the warnings about not existing files
            warnings.filterwarnings("ignore", message=".*path doesn't point to a file.*")
            if check_notna(row.get("file")):
                resource = _append_bitstream_to_resource(
                    resource=resource,
                    row=row,
                    row_number=row_number,
                )

    return resource


def _get_prop_function(
    row: pd.Series[Any],
    resource_id: str,
) -> Callable[..., etree._Element]:
    """
    Return the function that creates the appropriate property, depending on the proptype.

    Args:
        row: row of the CSV/Excel sheet that defines the property
        resource_id: resource ID of the resource to which the property belongs

    Raises:
        BaseError: if the proptype is invalid

    Returns:
        the function that creates the appropriate property
    """
    proptype_2_function: dict[str, Callable[..., etree._Element]] = {
        "bitstream": make_bitstream_prop,
        "boolean-prop": make_boolean_prop,
        "color-prop": make_color_prop,
        "date-prop": make_date_prop,
        "decimal-prop": make_decimal_prop,
        "geometry-prop": make_geometry_prop,
        "geoname-prop": make_geoname_prop,
        "integer-prop": make_integer_prop,
        "list-prop": make_list_prop,
        "resptr-prop": make_resptr_prop,
        "text-prop": make_text_prop,
        "uri-prop": make_uri_prop,
    }
    if row.get("prop type") not in proptype_2_function:
        raise BaseError(f"Invalid prop type for property {row.get('prop name')} in resource {resource_id}")
    return proptype_2_function[row["prop type"]]


def _convert_row_to_property_elements(
    row: pd.Series[Any],
    max_num_of_props: int,
    row_number: int,
    resource_id: str,
) -> list[PropertyElement]:
    """
    Every property contains i elements,
    which are represented in the Excel as groups of columns named
    {i_value, i_encoding, i_permissions, i_comment}.
    Depending on the property type, some of these cells are empty.
    This method converts a row to a list of PropertyElement objects.

    Args:
        row: the pandas series representing the current row
        max_num_of_props: highest number of properties that a resource in this file has
        row_number: row number of the CSV/Excel sheet
        resource_id: id of resource to which this property belongs to

    Warning:
        if a mandatory cell is missing, or if there are too many/too few values per property

    Returns:
        list of PropertyElement objects
    """
    property_elements: list[PropertyElement] = []
    for i in range(1, max_num_of_props + 1):
        value = row[f"{i}_value"]
        if pd.isna(value):
            # issue a warning if other cells of this property element are not empty
            # if all other cells are empty, continue with next property element
            other_cell_headers = [f"{i}_{x}" for x in ["encoding", "permissions", "comment"]]
            notna_cell_headers = [x for x in other_cell_headers if check_notna(row.get(x))]
            if notna_cell_headers_str := ", ".join([f"'{x}'" for x in notna_cell_headers]):
                warnings.warn(
                    f"Error in resource '{resource_id}': Excel row {row_number} has an entry "
                    f"in column(s) {notna_cell_headers_str}, but not in '{i}_value'. "
                    r"Please note that cell contents that don't meet the requirements of the regex [\p{L}\d_!?\-] "
                    "are considered inexistent."
                )
            continue

        # construct a PropertyElement from this property element
        kwargs_propelem = {"value": value, "permissions": str(row.get(f"{i}_permissions"))}
        if not check_notna(row.get(f"{i}_permissions")):
            warnings.warn(
                f"Resource '{resource_id}': "
                f"Missing permissions in column '{i}_permissions' of property '{row['prop name']}'"
            )
        if check_notna(row.get(f"{i}_comment")):
            kwargs_propelem["comment"] = str(row[f"{i}_comment"])
        if check_notna(row.get(f"{i}_encoding")):
            kwargs_propelem["encoding"] = str(row[f"{i}_encoding"])
        property_elements.append(PropertyElement(**kwargs_propelem))

    # validate the end result before returning it
    if not property_elements:
        warnings.warn(
            f"At least one value per property is required, "
            f"but resource '{resource_id}', property '{row['prop name']}' (Excel row {row_number}) doesn't contain any values."
        )
    if row.get("prop type") == "boolean-prop" and len(property_elements) > 1:
        warnings.warn(
            f"A <boolean-prop> can only have a single value, "
            f"but resource '{resource_id}', property '{row['prop name']}' (Excel row {row_number}) contains more than one value."
        )

    return property_elements


def _convert_property_row_to_xml(
    row_number: int,
    row: pd.Series[Any],
    max_num_of_props: int,
    resource_id: str,
) -> etree._Element:
    """
    Convert a property-row of the CSV/Excel sheet to an XML element.

    Args:
        row_number: row number of the CSV/Excel sheet
        row: the pandas series representing the current row
        max_num_of_props: highest number of properties that a resource in this file has
        resource_id: id of the resource to which the property will be appended

    Raises:
        BaseError: if there is inconsistent data in the row / if a validation fails

    Returns:
        the resource element with the appended property
    """
    # based on the property type, the right function has to be chosen
    make_prop_function = _get_prop_function(
        row=row,
        resource_id=resource_id,
    )

    # convert the row to a list of PropertyElement objects
    property_elements = _convert_row_to_property_elements(
        row=row,
        max_num_of_props=max_num_of_props,
        row_number=row_number,
        resource_id=resource_id,
    )

    # create the property
    return _create_property(
        make_prop_function=make_prop_function,
        row=row,
        property_elements=property_elements,
        resource_id=resource_id,
    )


def _create_property(
    make_prop_function: Callable[..., etree._Element],
    row: pd.Series[Any],
    property_elements: list[PropertyElement],
    resource_id: str,
) -> etree._Element:
    """
    Create a property based on the appropriate function and the property elements.

    Args:
        make_prop_function: the function to create the property
        row: the pandas series representing the current row of the Excel/CSV
        property_elements: the list of PropertyElement objects
        resource_id: id of resource to which this property belongs to

    Returns:
        the resource with the properties appended
    """
    kwargs_propfunc: dict[str, Union[str, PropertyElement, list[PropertyElement]]] = {
        "name": row["prop name"],
        "calling_resource": resource_id,
        "value": property_elements[0] if row.get("prop type") == "boolean-prop" else property_elements,
    }

    if check_notna(row.get("prop list")):
        kwargs_propfunc["list_name"] = str(row["prop list"])

    return make_prop_function(**kwargs_propfunc)


def excel2xml(
    datafile: str,
    shortcode: str,
    default_ontology: str,
) -> tuple[bool, list[warnings.WarningMessage]]:
    """
    This is a method that is called from the command line.
    It isn't intended to be used in a Python script.
    It takes a tabular data source in CSV/XLS(X) format that is formatted according to the specifications,
    and transforms it into a DSP-conforming XML file
    that can be uploaded to a DSP server with the xmlupload command.
    The output file is saved in the current working directory,
    with the name [default_ontology]-data.xml.

    Please note that this method doesn't do any data cleaning or data transformation tasks.
    The input and the output of this method are semantically exactly equivalent.

    Args:
        datafile: path to the data file (CSV or XLS(X))
        shortcode: shortcode of the project that this data belongs to
        default_ontology: name of the ontology that this data belongs to

    Raises:
        BaseError: if something went wrong

    Returns:
        True, empty list if everything went well,
        False and a list with caught warnings if there were any problems
    """
    success = True
    dataframe = _read_cli_input_file(datafile)
    dataframe = _validate_and_prepare_cli_input_file(dataframe)
    last_column_title = str(list(dataframe)[-1])  # last column title, in the format "i_comment"
    max_num_of_props = int(last_column_title.split("_")[0])
    output_file = Path(f"{default_ontology}-data.xml")

    root = make_root(shortcode=shortcode, default_ontology=default_ontology)
    root = append_permissions(root)

    with warnings.catch_warnings(record=True) as catched_warnings:
        resources = _convert_rows_to_xml(
            dataframe=dataframe,
            max_num_of_props=max_num_of_props,
        )
        root.extend(resources)
        write_xml(root, output_file)
        if len(catched_warnings) > 0:
            success = False
            for warning in catched_warnings:
                print(f"WARNING: {warning.message}")

    if success:
        print(f"XML file successfully written to '{output_file}'")
    else:
        red = "\033[31m"
        end = "\033[0m"
        print(f"{red}Some problems occurred. The XML file was written to '{output_file}', but it might be corrupt{end}")

    return success, catched_warnings



from __future__ import annotations

from typing import Any

import rustworkx as rx

from dsp_tools.commands.xmlupload.stash.graph_models import Cost
from dsp_tools.commands.xmlupload.stash.graph_models import Edge
from dsp_tools.commands.xmlupload.stash.graph_models import InfoForGraph
from dsp_tools.commands.xmlupload.stash.graph_models import LinkValueLink
from dsp_tools.commands.xmlupload.stash.graph_models import StandOffLink


def generate_upload_order(info_for_graph: InfoForGraph) -> tuple[dict[str, list[str]], list[str]]:
    """
    Generates the upload order from the Info for the graph

    Args:
        info_for_graph: Info for the graph

    Returns:
        - A dictionary which maps the resources that have stashes to the UUIDs of the stashed links.
        - A list of resource IDs which gives the order in which the resources should be uploaded to DSP-API.
    """
    graph, node_to_id, edges = _make_graph(info_for_graph)
    stash_lookup, upload_order, _ = _generate_upload_order_from_graph(graph, node_to_id, edges)
    return stash_lookup, upload_order


def _make_graph(
    info_for_graph: InfoForGraph,
) -> tuple[rx.PyDiGraph[Any, Any], dict[int, str], list[Edge]]:
    """
    This function takes information about the resources of an XML file and links between them.
    From that it constructs a rustworkx directed graph.
    Resources are represented as nodes and links as edges.

    Args:
        info_for_graph: Information required to construct the graph

    Returns:
        - The rustworkx graph.
        - A dictionary that maps the rustworkx index number of the nodes to the original resource ID from the XML file.
        - A list with all the edges in the graph.
    """
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    nodes = [(id_, None, None) for id_ in info_for_graph.all_resource_ids]
    node_indices = list(graph.add_nodes_from(nodes))
    id_to_node = dict(zip(info_for_graph.all_resource_ids, node_indices))
    node_to_id = dict(zip(node_indices, info_for_graph.all_resource_ids))
    edges = [Edge(id_to_node[x.source_id], id_to_node[x.target_id], x) for x in info_for_graph.link_values]
    for xml in info_for_graph.standoff_links:
        edges.extend([Edge(id_to_node[xml.source_id], id_to_node[x], xml) for x in xml.target_ids])
    graph.add_edges_from([e.as_tuple() for e in edges])
    return graph, node_to_id, edges


def _generate_upload_order_from_graph(
    graph: rx.PyDiGraph[Any, Any],
    node_to_id: dict[int, str],
    edges: list[Edge],
) -> tuple[dict[str, list[str]], list[str], int]:
    """
    Generate the order in which the resources should be uploaded to the DSP-API based on the dependencies.

    Args:
        graph: graph
        node_to_id: mapping between indices of the graph nodes and original resource IDs from the XML file
        edges: edges in the graph (contains info about source node, target node, and link info)

    Returns:
        - A dictionary which maps the resources that have stashes to the UUIDs of the stashed links.
        - A list of resource IDs which gives the order in which the resources should be uploaded to DSP-API.
        - The number of links in the stash.
    """
    upload_order: list[str] = []
    stash_lookup: dict[str, list[str]] = {}
    node_indices = set(node_to_id.keys())
    leaf_nodes, remaining_node_indices = _remove_leaf_nodes(graph, node_to_id, node_indices)
    upload_order.extend(leaf_nodes)
    stash_counter = 0
    while remaining_node_indices:
        cycle = list(rx.digraph_find_cycle(graph))
        links_to_remove = _find_cheapest_outgoing_links(graph, cycle, edges)
        stash_counter += len(links_to_remove)
        _remove_edges_to_stash(
            graph=graph,
            edges_to_remove=links_to_remove,
            all_edges=edges,
            remaining_nodes=remaining_node_indices,
        )
        stash_lookup = _add_stash_to_lookup_dict(stash_lookup, [x.link_object for x in links_to_remove])
        leaf_nodes, remaining_node_indices = _remove_leaf_nodes(graph, node_to_id, remaining_node_indices)
        upload_order.extend(leaf_nodes)
    return stash_lookup, upload_order, stash_counter


def _remove_leaf_nodes(
    graph: rx.PyDiGraph[Any, Any],
    node_to_id: dict[int, str],
    node_indices: set[int],
) -> tuple[list[str], set[int]]:
    """
    Leaf nodes are nodes that do not have any outgoing links.
    This means that they have no dependencies and are ok to upload.
    This function removes them from the graph.

    Args:
        graph: graph
        node_to_id: mapping of the rustworkx index number of the nodes to the original resource ID from the XML file
        node_indices: node indices that are in the graph

    Returns:
        - A list with the IDs of the removed leaf nodes.
        - A set with the indices of the nodes that remain in the graph.
    """
    removed_leaf_nodes: list[str] = []
    remaining_node_indices = set(node_indices)
    while leaf_nodes := [x for x in remaining_node_indices if graph.out_degree(x) == 0]:
        removed_leaf_nodes.extend(node_to_id[n] for n in leaf_nodes)
        graph.remove_nodes_from(leaf_nodes)
        remaining_node_indices = remaining_node_indices - set(leaf_nodes)
    return removed_leaf_nodes, remaining_node_indices


def _find_cheapest_outgoing_links(
    graph: rx.PyDiGraph[Any, Any],
    cycle: list[tuple[int, int]],
    edges: list[Edge],
) -> list[Edge]:
    """
    This function searches for the nodes whose outgoing links should be removed in order to break the cycle.
    It calculates which links between the resources create the smallest stash.

    Args:
        graph: graph
        cycle: the list with (source, target) for each edge in the cycle
        edges: edges in the graph (contains info about source node, target node, and link info)

    Returns:
        The edges (i.e. links) that should be stashed (containing all the edges connecting the two nodes)
    """
    costs: list[Cost] = []
    for source, target in cycle:
        edges_in = graph.in_edges(source)
        node_gain = len(edges_in)
        edges_out = graph.out_edges(source)
        node_cost = sum(x[2].cost_links for x in edges_out)
        node_value = node_cost / node_gain
        costs.append(Cost(source, target, node_value))
    cheapest_cost = sorted(costs, key=lambda x: x.node_value)[0]
    return [x for x in edges if x.source == cheapest_cost.source and x.target == cheapest_cost.target]


def _remove_edges_to_stash(
    graph: rx.PyDiGraph[Any, Any],
    edges_to_remove: list[Edge],
    all_edges: list[Edge],
    remaining_nodes: set[int],
) -> None:
    """
    This function removes the edges from the graph in order to break a cycle.

    Args:
        graph: graph
        edges_to_remove: edges that should be removed
        all_edges: all edges in the original graph
        remaining_nodes: indices of the nodes in the graph
    """
    normal_edges_to_remove = [(x.source, x.target) for x in edges_to_remove]
    # if only one (source, target) is removed, it removes only one edge, not all
    # therefore we need as many entries in the list as there are edges between the source and the target

    phantom_edges_to_remove = []
    source, target = edges_to_remove[0].source, edges_to_remove[0].target
    for link_to_stash in [x.link_object for x in edges_to_remove]:
        if isinstance(link_to_stash, StandOffLink):
            phantom_edges_to_remove.extend(
                _find_phantom_xml_edges(source, target, all_edges, link_to_stash, remaining_nodes)
            )

    all_edges_to_remove = normal_edges_to_remove + phantom_edges_to_remove
    graph.remove_edges_from(all_edges_to_remove)


def _find_phantom_xml_edges(
    source_node_index: int,
    target_node_index: int,
    all_edges: list[Edge],
    xml_link_to_stash: StandOffLink,
    remaining_nodes: set[int],
) -> list[tuple[int, int]]:
    """
    If an edge that will be removed represents an XML link,
    the text value may contain further links to other resources.
    If we stash the XMLLink, then in the real data all links of that text value are stashed.
    So, these "phantom" links must be removed from the graph.
    This function identifies the edges that must be removed from the rx graph.

    Args:
        source_node_index: rustworkx index of source node
        target_node_index: rustworkx index of target node
        all_edges: all edges in the original graph
        xml_link_to_stash: XML link that will be stashed
        remaining_nodes: indices of all nodes in the graph

    Returns:
        edges (rustworkx indices of nodes) that represent the links in the original XML text
    """

    def check(x: Edge) -> bool:
        return all(
            (
                x.source == source_node_index,
                x.target != target_node_index,
                x.link_object == xml_link_to_stash,
                x.target in remaining_nodes,
                # the target could have been removed because it was a leaf node, so we must check if it is still there
            )
        )

    return [(x.source, x.target) for x in all_edges if check(x)]


def _add_stash_to_lookup_dict(
    stash_dict: dict[str, list[str]],
    links_to_stash: list[StandOffLink | LinkValueLink],
) -> dict[str, list[str]]:
    stash_list = [stash_link.link_uuid for stash_link in links_to_stash]
    # all stashed links have the same subject id, so we can just take the first one
    subj_id = links_to_stash[0].source_id
    if subj_id in stash_dict:
        stash_dict[subj_id].extend(stash_list)
    else:
        stash_dict[subj_id] = stash_list
    return stash_dict



import subprocess
import sys
from pathlib import Path

import tomlkit


def main() -> None:
    """Get current version, append ".postN" (N=no. of commits since last release), write back into pyproject.toml"""

    # get the latest release tag from git history (e.g. "v9.0.2")
    latest_tag = _make_sys_call("git describe --tags --abbrev=0")

    new_version = _calculate_new_version(latest_tag)

    _write_new_version_into_pyproject_toml(old_version=latest_tag[1:], new_version=new_version)


def _calculate_new_version(latest_tag: str) -> str:
    _check_if_on_main_branch()
    # no. of commits since last release (e.g. "5")
    commit_count_str = _make_sys_call(f"git rev-list --count main ^{latest_tag}")
    # if this CI run was triggered by a release, do nothing
    if commit_count_str == "0":
        sys.exit(0)
    # e.g. 5 -> 4 (N in ".postN" is zero-based)
    commit_count = int(commit_count_str) - 1
    # e.g. "9.0.2.post4" (no leading v in pyproject.toml)
    return f"{latest_tag[1:]}.post{commit_count}"


def _check_if_on_main_branch() -> None:
    if _make_sys_call("git rev-parse --abbrev-ref HEAD") != "main":
        msg = (
            "The 'git rev-list' command will only work as expected on the main branch. "
            "If this workflow is intentionally run on a branch other than 'main', "
            "run 'git rev-list' with 'HEAD' instead of 'main'. "
            "This should only be done for testing purposes. In production, this will yield an unintended commit count."
        )
        raise ValueError(msg)


def _write_new_version_into_pyproject_toml(old_version: str, new_version: str) -> None:
    pyproject_pth = Path("pyproject.toml")
    pyproject = tomlkit.parse(pyproject_pth.read_text(encoding="utf-8"))
    if pyproject["project"]["version"] != old_version:  # type: ignore[index]
        msg = f"Version in pyproject.toml is '{pyproject['project']['version']}', but expected {old_version}"  # type: ignore[index]
        raise ValueError(msg)
    pyproject["project"]["version"] = new_version  # type: ignore[index]
    pyproject_pth.write_text(tomlkit.dumps(pyproject), encoding="utf-8")


def _make_sys_call(call: str) -> str:
    try:
        res = subprocess.run(call.split(), capture_output=True, check=True, encoding="utf-8").stdout.removesuffix("\n")
    except subprocess.CalledProcessError as e:
        print(f"ERROR: Call: {call} | Returncode: {e.returncode} | Stdout: {e.stdout} | Stderr: {e.stderr}")
        sys.exit(1)
    print(f"INFO: Call: {call} | Stdout: {res}")
    return res


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass

from rdflib import Literal
from rdflib import URIRef


@dataclass
class TransformedValue:
    value: Literal | URIRef
    prop_name: URIRef
    permissions: str | None
    comment: str | None


@dataclass
class Interval:
    start: Literal
    end: Literal


@dataclass
class RDFPropTypeInfo:
    knora_type: URIRef
    knora_prop: URIRef
    xsd_type: URIRef | None = None


@dataclass
class AbstractFileValue:
    value: str
    metadata: FileValueMetadata


@dataclass
class FileValueMetadata:
    license_iri: str | None
    copyright_holder: str | None
    authorships: list[str] | None
    permissions: str | None



import pytest
from lxml import etree

from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.values import BooleanValue
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import DateValue
from dsp_tools.xmllib.models.values import DecimalValue
from dsp_tools.xmllib.models.values import GeonameValue
from dsp_tools.xmllib.models.values import IntValue
from dsp_tools.xmllib.models.values import LinkValue
from dsp_tools.xmllib.models.values import ListValue
from dsp_tools.xmllib.models.values import Richtext
from dsp_tools.xmllib.models.values import SimpleText
from dsp_tools.xmllib.models.values import TimeValue
from dsp_tools.xmllib.models.values import UriValue
from dsp_tools.xmllib.models.values import Value
from dsp_tools.xmllib.serialise.serialise_values import serialise_values


def test_boolean() -> None:
    v: list[Value] = [BooleanValue("0", ":booleanProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<boolean-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":booleanProp">'
        b'<boolean permissions="open">false</boolean>'
        b"</boolean-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_color() -> None:
    v: list[Value] = [ColorValue("#000000", ":colorProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<color-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":colorProp">'
        b'<color permissions="open">#000000</color>'
        b"</color-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_date() -> None:
    v: list[Value] = [DateValue("2023-01-01", ":dateProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<date-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":dateProp">'
        b'<date permissions="open">2023-01-01</date>'
        b"</date-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_decimal() -> None:
    v: list[Value] = [DecimalValue("3.14", ":decimalProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<decimal-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":decimalProp">'
        b'<decimal permissions="open">3.14</decimal>'
        b"</decimal-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_geoname() -> None:
    v: list[Value] = [GeonameValue("99", ":geonameProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<geoname-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":geonameProp">'
        b'<geoname permissions="open">99</geoname>'
        b"</geoname-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_int() -> None:
    v: list[Value] = [IntValue("42", ":intProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<integer-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":intProp">'
        b'<integer permissions="open">42</integer>'
        b"</integer-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_link() -> None:
    v: list[Value] = [LinkValue("res_link", ":linkProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<resptr-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":linkProp">'
        b'<resptr permissions="open">res_link</resptr>'
        b"</resptr-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_list() -> None:
    v: list[Value] = [ListValue("item1", "listName", ":listProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<list-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":listProp" list="listName">'
        b'<list permissions="open">item1</list>'
        b"</list-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_richtext() -> None:
    v: list[Value] = [
        Richtext(
            "<otherTag>Hello World</otherTag>", ":richtextProp", resource_id="res_id", permissions=Permissions.OPEN
        )
    ]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<text-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":richtextProp">'
        b'<text permissions="open" encoding="xml">&lt;otherTag&gt;Hello World&lt;/otherTag&gt;</text>'
        b"</text-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_simpletext() -> None:
    v: list[Value] = [SimpleText("Hello World", ":simpleTextProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<text-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":simpleTextProp">'
        b'<text permissions="open" encoding="utf8">Hello World</text>'
        b"</text-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_time() -> None:
    v: list[Value] = [
        TimeValue("2009-10-10T12:00:00-05:00", ":timeProp", resource_id="res_id", permissions=Permissions.OPEN)
    ]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<time-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":timeProp">'
        b'<time permissions="open">2009-10-10T12:00:00-05:00</time>'
        b"</time-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_uri() -> None:
    v: list[Value] = [UriValue("https://example.com", ":uriProp", resource_id="res_id", permissions=Permissions.OPEN)]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<uri-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":uriProp">'
        b'<uri permissions="open">https://example.com</uri>'
        b"</uri-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_value_with_default_permission() -> None:
    v: list[Value] = [BooleanValue("0", ":booleanProp", resource_id="res_id")]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b"<boolean-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":booleanProp">'
        b"<boolean>false</boolean>"
        b"</boolean-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_several_values_different_generic_property() -> None:
    v: list[Value] = [
        LinkValue("res_link", ":linkProp", resource_id="res_id"),
        SimpleText("Hello World", ":simpleTextProp", resource_id="res_id"),
    ]
    result = serialise_values(v)
    assert len(result) == 2
    expected = (
        b"<resptr-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":linkProp"><resptr>res_link</resptr>'
        b"</resptr-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected
    expected = (
        b"<text-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":simpleTextProp">'
        b'<text encoding="utf8">Hello World</text>'
        b"</text-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_several_values_different_property() -> None:
    v: list[Value] = [
        BooleanValue("0", ":booleanProp", resource_id="res_id"),
        LinkValue("res_link", ":linkProp", resource_id="res_id"),
    ]
    result = serialise_values(v)
    assert len(result) == 2
    expected = (
        b'<boolean-prop xmlns="https://dasch.swiss/schema" '
        b'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":booleanProp">'
        b"<boolean>false</boolean>"
        b"</boolean-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected
    expected = (
        b"<resptr-prop "
        b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":linkProp">'
        b"<resptr>res_link</resptr>"
        b"</resptr-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


def test_several_values_same_property() -> None:
    v: list[Value] = [
        LinkValue("open_permission", ":linkProp", resource_id="res_id", permissions=Permissions.OPEN),
        LinkValue("default_permission", ":linkProp", resource_id="res_id"),
    ]
    result = serialise_values(v)
    assert len(result) == 1
    expected = (
        b'<resptr-prop xmlns="https://dasch.swiss/schema" '
        b'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'name=":linkProp">'
        b'<resptr permissions="open">open_permission</resptr>'
        b"<resptr>default_permission</resptr>"
        b"</resptr-prop>"
    )
    res_str = etree.tostring(result.pop(0))
    assert res_str == expected


if __name__ == "__main__":
    pytest.main([__file__])



from dsp_tools.commands.xmlupload.models.intermediary.res import IntermediaryResource
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.stash.graph_models import InfoForGraph
from dsp_tools.commands.xmlupload.stash.graph_models import LinkValueLink
from dsp_tools.commands.xmlupload.stash.graph_models import StandOffLink
from dsp_tools.utils.data_formats.iri_util import is_resource_iri


def create_info_for_graph_from_intermediary_resources(resources: list[IntermediaryResource]) -> InfoForGraph:
    """Extracts information to create the graph to analyse the circular references."""
    all_links = []
    all_stand_off = []
    all_resource_ids = []
    for res in resources:
        links, stand_off = _process_one_resource(res)
        all_links.extend(links)
        all_stand_off.extend(stand_off)
        all_resource_ids.append(res.res_id)
    return InfoForGraph(
        all_resource_ids=all_resource_ids,
        link_values=all_links,
        standoff_links=all_stand_off,
    )


def _process_one_resource(resource: IntermediaryResource) -> tuple[list[LinkValueLink], list[StandOffLink]]:
    link_values = []
    stand_off = []
    for val in resource.values:
        if isinstance(val, IntermediaryLink):
            if is_resource_iri(val.value):
                continue
            link_values.append(
                LinkValueLink(
                    source_id=resource.res_id,
                    target_id=val.value,
                    link_uuid=val.value_uuid,
                )
            )
        elif isinstance(val, IntermediaryRichtext):
            if val.resource_references:
                only_ids = {x for x in val.resource_references if not is_resource_iri(x)}
                if not only_ids:
                    continue
                stand_off.append(
                    StandOffLink(
                        source_id=resource.res_id,
                        target_ids=only_ids,
                        link_uuid=val.value_uuid,
                    )
                )
    return link_values, stand_off



from loguru import logger
from rdflib import Graph


def construct_allowed_licenses_shape() -> Graph:
    """Create a constraint detailing the allowed licences."""
    logger.info("Constructing allowed licesnses shapes.")
    ttl_str = """
    @prefix sh:         <http://www.w3.org/ns/shacl#> .
    @prefix knora-api:  <http://api.knora.org/ontology/knora-api/v2#> .
    @prefix api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#> .
    
    api-shapes:FileValue_ClassShape
      a sh:NodeShape ;
      sh:targetClass knora-api:FileValue ;
      sh:property [
          a sh:PropertyShape ;
          sh:path knora-api:hasLicense ;
          sh:in (
                  <http://rdfh.ch/licenses/cc-by-4.0> <http://rdfh.ch/licenses/cc-by-sa-4.0>
                  <http://rdfh.ch/licenses/cc-by-nc-4.0> <http://rdfh.ch/licenses/cc-by-nc-sa-4.0>
                  <http://rdfh.ch/licenses/cc-by-nd-4.0> <http://rdfh.ch/licenses/cc-by-nc-nd-4.0>
                  <http://rdfh.ch/licenses/ai-generated> <http://rdfh.ch/licenses/unknown>
                  <http://rdfh.ch/licenses/public-domain>
                ) ;
          sh:message "You are required to use one of the pre-defined licenses, please consult the documentation for details." ;
          sh:severity sh:Violation
                  ] .
    """  # noqa: E501 Line too long (135 > 120)
    g = Graph()
    g.parse(data=ttl_str, format="turtle")
    return g



from __future__ import annotations

import json
import os
import time
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from typing import Any
from typing import Literal
from typing import Never
from typing import Union

from loguru import logger
from requests import JSONDecodeError
from requests import ReadTimeout
from requests import Response

from dsp_tools.commands.project.legacy_models.context import Context
from dsp_tools.commands.project.legacy_models.helpers import OntoIri
from dsp_tools.error.exceptions import PermanentTimeOutError


@dataclass
class PostFiles:
    """One or more files to be uploaded in a POST request."""

    files: list[PostFile]

    def to_dict(self) -> dict[str, tuple[str, Any, str] | tuple[str, Any]]:
        return {x.file_name: x.to_tuple() for x in self.files}


@dataclass
class PostFile:
    file_name: str
    fileobj: Any
    content_type: str | None = None

    def to_tuple(self) -> tuple[str, Any, str] | tuple[str, Any]:
        if self.content_type:
            return self.file_name, self.fileobj, self.content_type
        return self.file_name, self.fileobj


class SetEncoder(json.JSONEncoder):
    """Encoder used to serialize objects to JSON that would by default not be serializable"""

    def default(self, o: Union[set[Any], Context, OntoIri]) -> Any:
        """Return a serializable object for o"""
        if isinstance(o, set):
            return list(o)
        elif isinstance(o, Context):
            return o.toJsonObj()
        elif isinstance(o, OntoIri):
            return {"iri": o.iri, "hashtag": o.hashtag}
        return json.JSONEncoder.default(self, o)


@dataclass
class RequestParameters:
    method: Literal["POST", "GET", "PUT", "DELETE"]
    url: str
    timeout: int
    data: dict[str, Any] | None = None
    data_serialized: bytes | None = field(init=False, default=None)
    headers: dict[str, str] | None = None
    files: PostFiles | None = None

    def __post_init__(self) -> None:
        self.data_serialized = self._serialize_payload(self.data)

    def _serialize_payload(self, payload: dict[str, Any] | None) -> bytes | None:
        # If data is not encoded as bytes, issues can occur with non-ASCII characters,
        # where the content-length of the request will turn out to be different from the actual length.
        return json.dumps(payload, cls=SetEncoder, ensure_ascii=False).encode("utf-8") if payload else None

    def as_kwargs(self) -> dict[str, Any]:
        kwargs = {
            "method": self.method,
            "url": self.url,
            "timeout": self.timeout,
            "data": self.data_serialized,
            "headers": self.headers,
        }
        if self.files:
            kwargs["files"] = self.files.to_dict()
        return kwargs


def log_request(params: RequestParameters, extra_headers: dict[str, Any] | None = None) -> None:
    """Logs the request."""
    dumpobj = {
        "method": params.method,
        "url": params.url,
        "timeout": params.timeout,
    }
    headers_to_log = {}
    if extra_headers:
        headers_to_log = extra_headers
    if params.headers:
        headers_to_log = headers_to_log | params.headers
    dumpobj["headers"] = sanitize_headers(headers_to_log)
    if params.data:
        data = params.data.copy()
        if "password" in data:
            data["password"] = "***"
        dumpobj["data"] = data
    if params.files:
        dumpobj["files"] = [x.file_name for x in params.files.files]
    logger.debug(f"REQUEST: {json.dumps(dumpobj, cls=SetEncoder)}")


def log_response(response: Response, include_response_content: bool = True) -> None:
    """Log the response of a request."""
    dumpobj: dict[str, Any] = {
        "status_code": response.status_code,
        "headers": sanitize_headers(dict(response.headers)) if response.headers else "",
    }
    if include_response_content:
        try:
            dumpobj["content"] = response.json()
        except JSONDecodeError:
            dumpobj["content"] = response.text
    else:
        dumpobj["content"] = "too big to be logged"
    logger.debug(f"RESPONSE: {json.dumps(dumpobj)}")


def sanitize_headers(headers: dict[str, str | bytes]) -> dict[str, str]:
    """Remove sensitive information from request headers."""

    def _mask(key: str, value: str | bytes) -> str:
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        if key == "Authorization" and value.startswith("Bearer "):
            return "Bearer ***"
        if key == "Set-Cookie":
            return "***"
        return value

    return {k: _mask(k, v) for k, v in headers.items()}


def log_request_failure_and_sleep(reason: str, retry_counter: int, exc_info: bool) -> None:
    """Log the reason for a request failure and sleep."""
    msg = f"{reason}: Try reconnecting to DSP server, next attempt in {2**retry_counter} seconds..."
    print(f"{datetime.now()}: {msg}")
    if exc_info:
        logger.exception(f"{msg} ({retry_counter=:})")
    else:
        logger.error(f"{msg} ({retry_counter=:})")
    time.sleep(2**retry_counter)


def log_and_raise_timeouts(error: TimeoutError | ReadTimeout) -> Never:
    """Log a timeout error raised by a request and raise our own PermanentTimeOutError"""
    msg = f"A '{error.__class__.__name__}' occurred during the connection to the DSP server."
    print(f"{datetime.now()}: {msg}")
    logger.error(msg)
    raise PermanentTimeOutError(msg) from None


def should_retry(response: Response) -> bool:
    """Returns the decision if a retry of a request is sensible."""
    in_500_range = 500 <= response.status_code < 600
    try_again_later = "try again later" in response.text.lower()
    in_testing_env = os.getenv("DSP_TOOLS_TESTING") == "true"  # set in .github/workflows/tests-on-push.yml
    return (try_again_later or in_500_range) and not in_testing_env



import pytest

from dsp_tools.commands.xmlupload.upload_config import _transform_server_url_to_foldername


class TestTransformServerUrlToFoldername:
    def test_prod_env(self) -> None:
        servername = "https://api.dasch.swiss/"
        expected = "dasch.swiss"
        result = _transform_server_url_to_foldername(servername)
        assert result == expected

    def test_test_env(self) -> None:
        servername = "https://api.test.dasch.swiss/"
        expected = "test.dasch.swiss"
        result = _transform_server_url_to_foldername(servername)
        assert result == expected

    def test_project_server(self) -> None:
        servername = "http://api.082e-test-server.dasch.swiss/"
        expected = "082e-test-server.dasch.swiss"
        result = _transform_server_url_to_foldername(servername)
        assert result == expected

    def test_local_http(self) -> None:
        servername = "http://0.0.0.0:12345"
        expected = "localhost"
        result = _transform_server_url_to_foldername(servername)
        assert result == expected

    def test_local_https(self) -> None:
        servername = "https://0.0.0.0:80/"
        expected = "localhost"
        result = _transform_server_url_to_foldername(servername)
        assert result == expected


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeAlias
from typing import Union

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.utils.data_formats.date_util import Date


@dataclass
class IntervalFloats:
    start: float
    end: float


IntermediaryValueTypes: TypeAlias = Union[bool, str, float, int, FormattedTextValue, Date, IntervalFloats]


@dataclass
class IntermediaryValue:
    value: IntermediaryValueTypes
    prop_iri: str
    comment: str | None
    permissions: Permissions | None


@dataclass
class IntermediaryBoolean(IntermediaryValue):
    value: bool


@dataclass
class IntermediaryColor(IntermediaryValue):
    value: str


@dataclass
class IntermediaryDate(IntermediaryValue):
    value: Date


@dataclass
class IntermediaryDecimal(IntermediaryValue):
    value: float


@dataclass
class IntermediaryGeoname(IntermediaryValue):
    value: str


@dataclass
class IntermediaryGeometry(IntermediaryValue):
    value: str


@dataclass
class IntermediaryInt(IntermediaryValue):
    value: int


@dataclass
class IntermediaryInterval(IntermediaryValue):
    value: IntervalFloats


@dataclass
class IntermediaryLink(IntermediaryValue):
    value: str
    value_uuid: str

    def __post_init__(self) -> None:
        self.prop_iri = f"{self.prop_iri}Value"


@dataclass
class IntermediaryList(IntermediaryValue):
    value: str


@dataclass
class IntermediarySimpleText(IntermediaryValue):
    value: str


@dataclass
class IntermediaryRichtext(IntermediaryValue):
    value: FormattedTextValue
    resource_references: set[str]
    value_uuid: str


@dataclass
class IntermediaryTime(IntermediaryValue):
    value: str


@dataclass
class IntermediaryUri(IntermediaryValue):
    value: str



from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

import pandas as pd

separator = "\n    "
list_separator = "\n    - "


@dataclass(frozen=True)
class IngestInformation:
    """
    This class stores the information about the mapping of ids provided by the dsp-ingest service
    and the filepaths used in the XML file.
    """

    unused_mediafiles: list[str]
    mediafiles_no_id: list[tuple[str, str]]
    maximum_prints: int = 20
    csv_directory_path: Path = field(default=Path.cwd())
    unused_mediafiles_csv: str = "UnusedMediaUploaded.csv"
    mediafiles_no_id_csv: str = "FilesNotUploaded.csv"

    def ok_msg(self) -> str | None:
        """
        This function checks if no media was unused or not uploaded.
        If that is the case it returns the message,
        if not, it ends without an effect.

        Returns:
            Message if all went well.
        """
        if not self.unused_mediafiles and not self.mediafiles_no_id:
            return (
                "All multimedia files referenced in the XML file were uploaded through dsp-ingest.\n"
                "All multimedia files uploaded through dsp-ingest were referenced in the XML file."
            )
        return None

    def execute_error_protocol(self) -> str:
        """
        This function generates the user message and saves a file with the information
        if a lot of resources are affected.

        Returns:
            User message
        """
        self._save_csv_if_applicable()
        return self._get_error_msg()

    def _get_error_msg(self) -> str:
        msg_list = [
            "The upload cannot continue as there are problems with the multimedia files referenced in the XML.",
        ]
        if has_msg := self._get_unused_mediafiles_msg():
            msg_list.append(has_msg)
        if has_msg := self._get_mediafiles_no_id_msg():
            msg_list.append(has_msg)
        return separator.join(msg_list)

    def _get_mediafiles_no_id_msg(self) -> str | None:
        if 0 < len(self.mediafiles_no_id) <= self.maximum_prints:
            return (
                "The data XML file contains references to the following multimedia files "
                "which were not previously uploaded through dsp-ingest:"
                + list_separator
                + list_separator.join([f"Resource ID: '{x[0]}' | Filepath: '{x[1]}'" for x in self.mediafiles_no_id])
            )
        elif len(self.mediafiles_no_id) > self.maximum_prints:
            return (
                "The data XML file contains references to multimedia files "
                "which were not previously uploaded through dsp-ingest:\n"
                f"    The file with the resource IDs and problematic filenames was saved at "
                f"'{Path(self.csv_directory_path / self.mediafiles_no_id_csv)}'."
            )
        return None

    def _get_unused_mediafiles_msg(self) -> str | None:
        if 0 < len(self.unused_mediafiles) <= self.maximum_prints:
            return (
                "The data XML file does not reference the following multimedia files which were previously "
                "uploaded through dsp-ingest:" + list_separator + list_separator.join(self.unused_mediafiles)
            )
        elif len(self.unused_mediafiles) > self.maximum_prints:
            return (
                "The data XML file does not reference all the multimedia files which were previously "
                "uploaded through dsp-ingest.\n"
                f"    The file with the unused filenames was saved at "
                f"'{Path(self.csv_directory_path / self.unused_mediafiles_csv)}'."
            )
        return None

    def _save_csv_if_applicable(self) -> None:
        if (unused_mediafiles_df := self._unused_mediafiles_to_df()) is not None:
            _save_as_csv(unused_mediafiles_df, self.csv_directory_path, self.unused_mediafiles_csv)
        if (no_id_df := self._mediafiles_no_id_to_df()) is not None:
            _save_as_csv(no_id_df, self.csv_directory_path, self.mediafiles_no_id_csv)

    def _unused_mediafiles_to_df(self) -> pd.DataFrame | None:
        return (
            pd.DataFrame({"Multimedia Filenames": self.unused_mediafiles})
            if len(self.unused_mediafiles) > self.maximum_prints
            else None
        )

    def _mediafiles_no_id_to_df(self) -> pd.DataFrame | None:
        return (
            pd.DataFrame(
                {
                    "Resource ID": [x[0] for x in self.mediafiles_no_id],
                    "Filepath": [x[1] for x in self.mediafiles_no_id],
                }
            )
            if len(self.mediafiles_no_id) > self.maximum_prints
            else None
        )


def _save_as_csv(df: pd.DataFrame, directory_path: Path, filename: str) -> None:
    df.to_csv(Path(directory_path, filename), index=False)



import unittest
from typing import Callable

import regex
from lxml import etree

from dsp_tools.commands import excel2xml
from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.error.exceptions import BaseError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)
# ruff: noqa: PT027 (pytest-unittest-raises-assertion) (remove this line when pytest is used instead of unittest)


class TestResourceTypes(unittest.TestCase):
    def test_make_link(self) -> None:
        expected = '<link label="label" id="id" permissions="open"/>'
        result = _strip_namespace(excel2xml.make_link("label", "id"))
        self.assertEqual(expected, result)

    def test_make_link_with_permission(self) -> None:
        expected = '<link label="label" id="id" permissions="restricted"/>'
        result = _strip_namespace(excel2xml.make_link("label", "id", "restricted"))
        self.assertEqual(expected, result)

    def test_make_link_with_ark(self) -> None:
        expected = '<link label="label" id="id" permissions="open" ark="ark"/>'
        result = _strip_namespace(excel2xml.make_link("label", "id", ark="ark"))
        self.assertEqual(expected, result)

    def test_make_link_with_iri(self) -> None:
        expected = '<link label="label" id="id" permissions="open" iri="iri"/>'
        result = _strip_namespace(excel2xml.make_link("label", "id", iri="iri"))
        self.assertEqual(expected, result)

    def test_make_link_with_creation_date(self) -> None:
        expected = '<link label="label" id="id" permissions="open" creation_date="2019-10-23T13:45:12Z"/>'
        result = _strip_namespace(excel2xml.make_link("label", "id", creation_date="2019-10-23T13:45:12Z"))
        self.assertEqual(expected, result)

    def test_warn_make_link_with_iri_and_ark(self) -> None:
        with self.assertWarns(DspToolsUserWarning):
            excel2xml.make_link("label", "id", ark="ark", iri="iri")

    def test_fail_link_with_invalid_creation_date(self) -> None:
        with self.assertRaisesRegex(BaseError, "invalid creation date"):
            excel2xml.make_link("label", "id", creation_date="2019-10-23T13:45:12")

    def test_make_region(self) -> None:
        expected = '<region label="label" id="id" permissions="open"/>'
        result = _strip_namespace(excel2xml.make_region("label", "id"))
        self.assertEqual(expected, result)

    def test_make_region_with_permission(self) -> None:
        expected = '<region label="label" id="id" permissions="restricted"/>'
        result = _strip_namespace(excel2xml.make_region("label", "id", "restricted"))
        self.assertEqual(expected, result)

    def test_make_region_with_ark(self) -> None:
        expected = '<region label="label" id="id" permissions="open" ark="ark"/>'
        result = _strip_namespace(excel2xml.make_region("label", "id", ark="ark"))
        self.assertEqual(expected, result)

    def test_make_region_with_iri(self) -> None:
        expected = '<region label="label" id="id" permissions="open" iri="iri"/>'
        result = _strip_namespace(excel2xml.make_region("label", "id", iri="iri"))
        self.assertEqual(expected, result)

    def test_make_region_with_creation_date(self) -> None:
        expected = '<region label="label" id="id" permissions="open" creation_date="2019-10-23T13:45:12Z"/>'
        result = _strip_namespace(excel2xml.make_region("label", "id", creation_date="2019-10-23T13:45:12Z"))
        self.assertEqual(expected, result)

    def test_warn_make_region_with_iri_and_ark(self) -> None:
        with self.assertWarns(DspToolsUserWarning):
            excel2xml.make_region("label", "id", ark="ark", iri="iri")

    def test_fail_region_with_invalid_creation_date(self) -> None:
        with self.assertRaisesRegex(BaseError, "invalid creation date"):
            excel2xml.make_region("label", "id", creation_date="2019-10-23T13:45:12")

    def test_make_audio_segment(self) -> None:
        expected = '<audio-segment label="label" id="id" permissions="open"/>'
        result = _strip_namespace(excel2xml.make_audio_segment("label", "id"))
        self.assertEqual(expected, result)

    def test_make_audio_segment_with_custom_permissions(self) -> None:
        expected = '<audio-segment label="label" id="id" permissions="restricted"/>'
        result = _strip_namespace(excel2xml.make_audio_segment("label", "id", "restricted"))
        self.assertEqual(expected, result)

    def test_make_video_segment(self) -> None:
        expected = '<video-segment label="label" id="id" permissions="open"/>'
        result = _strip_namespace(excel2xml.make_video_segment("label", "id"))
        self.assertEqual(expected, result)

    def test_make_video_segment_with_custom_permissions(self) -> None:
        expected = '<video-segment label="label" id="id" permissions="restricted"/>'
        result = _strip_namespace(excel2xml.make_video_segment("label", "id", "restricted"))
        self.assertEqual(expected, result)

    def test_make_resource(self) -> None:
        test_cases: list[tuple[Callable[..., etree._Element], str]] = [
            (
                lambda: excel2xml.make_resource("label", "restype", "id"),
                '<resource label="label" restype="restype" id="id" permissions="open"/>',
            ),
            (
                lambda: excel2xml.make_resource("label", "restype", "id", "restricted"),
                '<resource label="label" restype="restype" id="id" permissions="restricted"/>',
            ),
            (
                lambda: excel2xml.make_resource("label", "restype", "id", ark="ark"),
                '<resource label="label" restype="restype" id="id" permissions="open" ark="ark"/>',
            ),
            (
                lambda: excel2xml.make_resource("label", "restype", "id", iri="iri"),
                '<resource label="label" restype="restype" id="id" permissions="open" iri="iri"/>',
            ),
            (
                lambda: excel2xml.make_resource("label", "restype", "id", creation_date="2019-10-23T13:45:12Z"),
                (
                    '<resource label="label" restype="restype" id="id" permissions="open" '
                    'creation_date="2019-10-23T13:45:12Z"/>'
                ),
            ),
        ]

        for method, result in test_cases:
            xml_returned_as_element = method()
            xml_returned = etree.tostring(xml_returned_as_element, encoding="unicode")
            xml_returned = regex.sub(r" xmlns(:.+?)?=\".+?\"", "", xml_returned)
            self.assertEqual(result, xml_returned)

        self.assertWarns(
            DspToolsUserWarning, lambda: excel2xml.make_resource("label", "restype", "id", ark="ark", iri="iri")
        )
        with self.assertRaisesRegex(BaseError, "invalid creation date"):
            excel2xml.make_resource("label", "restype", "id", creation_date="2019-10-23T13:45:12")


def _strip_namespace(element: etree._Element) -> str:
    """Removes the namespace from the XML element."""
    xml = etree.tostring(element, encoding="unicode")
    xml = regex.sub(r" xmlns(:.+?)?=\".+?\"", "", xml)
    return xml



import warnings

import pytest

from dsp_tools.xmllib.models.geometry import Circle
from dsp_tools.xmllib.models.geometry import GeometryPoint
from dsp_tools.xmllib.models.geometry import Polygon
from dsp_tools.xmllib.models.geometry import Rectangle
from dsp_tools.xmllib.models.geometry import Vector


class TestGeometryShape:
    def test_rectangle_to_json_string_success(self) -> None:
        geom_obj = Rectangle(
            point1=GeometryPoint(0.1, 0.2, "res_id"),
            point2=GeometryPoint(0.2, 0.3, "res_id"),
            line_width=2,
            color="#5b24bf",
            active=True,
            resource_id="res_id",
        )
        expected_str = (
            '{"status": "active", "type": "rectangle", "lineWidth": 2, '
            '"points": [{"x": 0.1, "y": 0.2}, {"x": 0.2, "y": 0.3}]}'
        )
        result = geom_obj.to_json_string()
        assert result == expected_str

    def test_polygon_to_json_string_no_points(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            polygon = Polygon(
                points=[],
                line_width=2,
                color="#5b24bf",
                active=True,
                resource_id="res_id",
            )
            assert len(caught_warnings) == 1
        result = polygon.to_json_string()
        expected_str = '{"status": "active", "type": "polygon", "lineWidth": 2, "points": []}'
        assert result == expected_str

    def test_circle_to_json_string_success(self) -> None:
        geom_obj = Circle(
            center=GeometryPoint(0.1, 0.2, "res_id"),
            radius=Vector(0.2, 0.2, "res_id"),
            line_width=2,
            color="#5b24bf",
            active=True,
            resource_id="res_id",
        )
        expected_str = (
            '{"status": "active", "type": "circle", "lineWidth": 2, '
            '"points": [{"x": 0.1, "y": 0.2}], "radius": {"x": 0.2, "y": 0.2}}'
        )
        result = geom_obj.to_json_string()
        assert result == expected_str


class TestGeometryPoint:
    def test_success(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            point = GeometryPoint(0.1, 0.2, "res_id")
            assert len(caught_warnings) == 0
        assert point.x == 0.1
        assert point.y == 0.2

    def test_success_as_string(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            point = GeometryPoint("0.1", "0.2", "res_id")  # type: ignore[arg-type]
            assert len(caught_warnings) == 0
        assert point.x == 0.1
        assert point.y == 0.2

    def test_warning(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            point = GeometryPoint("x", "y", "res_id")  # type: ignore[arg-type]
            assert len(caught_warnings) == 1
        assert point.x == "x"  # type: ignore[comparison-overlap]
        assert point.y == "y"  # type: ignore[comparison-overlap]


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="method-assign,no-untyped-def"
import pytest
import regex

from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryFileValue
from dsp_tools.commands.xmlupload.models.intermediary.file_values import IntermediaryIIIFUri
from dsp_tools.commands.xmlupload.models.intermediary.res import MigrationMetadata
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryBoolean
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryColor
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDate
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryDecimal
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeometry
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryGeoname
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInt
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryInterval
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryLink
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryList
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryRichtext
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediarySimpleText
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryTime
from dsp_tools.commands.xmlupload.models.intermediary.values import IntermediaryUri
from dsp_tools.commands.xmlupload.models.lookup_models import IntermediaryLookups
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import _get_metadata
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import _resolve_permission
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import _transform_file_value
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import _transform_iiif_uri_value
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import _transform_one_resource
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import _transform_one_value
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_into_intermediary_classes import (
    transform_all_resources_into_intermediary_resources,
)
from dsp_tools.error.exceptions import InputError
from dsp_tools.error.exceptions import PermissionNotExistsError
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.utils.data_formats.date_util import Date
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedMigrationMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedValue

ONTO = "http://0.0.0.0:3333/ontology/9999/onto/v2#"
HAS_PROP = f"{ONTO}hasProp"
RES_TYPE = f"{ONTO}ResourceType"


@pytest.fixture
def lookups() -> IntermediaryLookups:
    return IntermediaryLookups(
        permissions={"open": Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})},
        listnodes={("list", "node"): "http://rdfh.ch/9999/node"},
        namespaces={
            "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
            "onto": "http://0.0.0.0:3333/ontology/9999/onto/v2#",
        },
        authorships={"auth_id": ["author"]},
    )


@pytest.fixture
def file_with_permission() -> ParsedFileValue:
    metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", "open")
    return ParsedFileValue("file.jpg", KnoraValueType.STILL_IMAGE_FILE, metadata)


@pytest.fixture
def bool_value() -> ParsedValue:
    return ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, None, None)


@pytest.fixture
def iiif_file_value():
    metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", None)
    return ParsedFileValue("https://this/is/a/uri.jpg", KnoraValueType.STILL_IMAGE_IIIF, metadata)


class TestTransformResources:
    def test_success(self, lookups):
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=None,
            migration_metadata=None,
        )
        result = transform_all_resources_into_intermediary_resources([res], lookups)
        assert len(result.transformed_resources) == 1
        assert not result.resource_failures

    def test_failure(self, lookups: IntermediaryLookups):
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id="nonExisting",
            values=[],
            file_value=None,
            migration_metadata=None,
        )
        result = transform_all_resources_into_intermediary_resources([res], lookups)
        assert not result.transformed_resources
        assert len(result.resource_failures) == 1
        assert result.resource_failures[0].resource_id == "id"
        assert result.resource_failures[0].failure_msg == "Could not find permissions for value: nonExisting"


class TestTransformOneResource:
    def test_resource_one_value(self, bool_value, lookups: IntermediaryLookups):
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[bool_value],
            file_value=None,
            migration_metadata=None,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 1
        assert not result.file_value
        assert not result.migration_metadata

    def test_resource_with_permissions(self, lookups: IntermediaryLookups):
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id="open",
            values=[],
            file_value=None,
            migration_metadata=None,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert isinstance(result.permissions, Permissions)
        assert len(result.values) == 0
        assert not result.file_value
        assert not result.migration_metadata

    def test_with_ark(self, lookups: IntermediaryLookups):
        parsed_metadata = ParsedMigrationMetadata(
            iri=None, ark="ark:/72163/4123-43xc6ivb931-a.2022829", creation_date="1999-12-31T23:59:59.9999999+01:00"
        )
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=None,
            migration_metadata=parsed_metadata,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        metadata = result.migration_metadata
        assert isinstance(metadata, MigrationMetadata)
        assert metadata.iri_str == "http://rdfh.ch/4123/5d5d1FKaUC2Wfl4zicggfg"
        assert metadata.creation_date == "1999-12-31T23:59:59.9999999+01:00"

    def test_with_iri(self, lookups: IntermediaryLookups):
        parsed_metadata = ParsedMigrationMetadata(
            iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA", ark=None, creation_date=None
        )
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=None,
            migration_metadata=parsed_metadata,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        metadata = result.migration_metadata
        assert isinstance(metadata, MigrationMetadata)
        assert metadata.iri_str == "http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA"
        assert not metadata.creation_date

    def test_resource_with_ark_and_iri(self, lookups: IntermediaryLookups):
        parsed_metadata = ParsedMigrationMetadata(
            iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
            ark="ark:/72163/4123-43xc6ivb931-a.2022829",
            creation_date="1999-12-31T23:59:59.9999999+01:00",
        )
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=None,
            migration_metadata=parsed_metadata,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        metadata = result.migration_metadata
        assert isinstance(metadata, MigrationMetadata)
        assert metadata.iri_str == "http://rdfh.ch/4123/5d5d1FKaUC2Wfl4zicggfg"
        time_stamp = metadata.creation_date
        assert isinstance(time_stamp, DateTimeStamp)
        assert time_stamp == DateTimeStamp("1999-12-31T23:59:59.9999999+01:00")

    def test_unknown_permission(self, lookups: IntermediaryLookups):
        msg = regex.escape(r"Could not find permissions for value: unknown")
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id="unknown",
            values=[],
            file_value=None,
            migration_metadata=None,
        )
        with pytest.raises(PermissionNotExistsError, match=msg):
            _transform_one_resource(res, lookups)

    def test_with_file_value(self, file_with_permission, lookups: IntermediaryLookups):
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=file_with_permission,
            migration_metadata=None,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        file_val = result.file_value
        assert isinstance(file_val, IntermediaryFileValue)
        assert file_val.value == "file.jpg"
        assert file_val.metadata.permissions
        assert not result.iiif_uri
        assert not result.migration_metadata

    def test_with_iiif_uri(self, iiif_file_value, lookups: IntermediaryLookups):
        res = ParsedResource(
            res_id="id",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=iiif_file_value,
            migration_metadata=None,
        )
        result = _transform_one_resource(res, lookups)
        assert result.res_id == "id"
        assert result.type_iri == RES_TYPE
        assert result.label == "lbl"
        assert not result.permissions
        assert len(result.values) == 0
        assert not result.file_value
        assert isinstance(result.iiif_uri, IntermediaryIIIFUri)
        assert not result.migration_metadata


class TestTransformFileValue:
    def test_transform_file_value(self, lookups: IntermediaryLookups):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", None)
        val = ParsedFileValue("file.jpg", KnoraValueType.STILL_IMAGE_FILE, metadata)
        result = _transform_file_value(val, lookups, "id", "lbl")
        assert result.value == "file.jpg"
        assert isinstance(result, IntermediaryFileValue)
        result_metadata = result.metadata
        assert not result_metadata.permissions
        assert result_metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert result_metadata.copyright_holder == "copy"
        assert result_metadata.authorships == ["author"]

    def test_transform_file_value_with_permissions(self, file_with_permission, lookups: IntermediaryLookups):
        result = _transform_file_value(file_with_permission, lookups, "id", "lbl")
        assert isinstance(result, IntermediaryFileValue)
        assert result.value == "file.jpg"
        result_metadata = result.metadata
        assert isinstance(result_metadata.permissions, Permissions)
        assert result_metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert result_metadata.copyright_holder == "copy"
        assert result_metadata.authorships == ["author"]

    def test_transform_iiif_uri_value(self, iiif_file_value, lookups: IntermediaryLookups):
        result = _transform_iiif_uri_value(iiif_file_value, lookups)
        assert result.value == "https://this/is/a/uri.jpg"
        assert isinstance(result, IntermediaryIIIFUri)
        result_metadata = result.metadata
        assert not result_metadata.permissions
        assert result_metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert result_metadata.copyright_holder == "copy"
        assert result_metadata.authorships == ["author"]

    def test_transform_iiif_uri_value_with_permission(self, lookups: IntermediaryLookups):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", "open")
        val = ParsedFileValue("https://this/is/a/uri.jpg", KnoraValueType.STILL_IMAGE_FILE, metadata)
        result = _transform_iiif_uri_value(val, lookups)
        assert isinstance(result, IntermediaryIIIFUri)
        assert result.value == "https://this/is/a/uri.jpg"
        result_metadata = result.metadata
        assert isinstance(result_metadata.permissions, Permissions)
        assert result_metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert result_metadata.copyright_holder == "copy"
        assert result_metadata.authorships == ["author"]

    def test_get_metadata_soon_deprecated_without_metadata(self, lookups):
        metadata = ParsedFileValueMetadata(None, None, None, None)
        val = ParsedFileValue("https://this/is/a/uri.jpg", KnoraValueType.STILL_IMAGE_FILE, metadata)
        result = _transform_iiif_uri_value(val, lookups)
        assert not result.metadata.permissions
        assert not result.metadata.license_iri
        assert not result.metadata.copyright_holder
        assert not result.metadata.authorships


class TestFileMetadata:
    def test_good(self, lookups):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", "open")
        result_metadata = _get_metadata(metadata, lookups)
        assert isinstance(result_metadata.permissions, Permissions)
        assert result_metadata.license_iri == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert result_metadata.copyright_holder == "copy"
        assert result_metadata.authorships == ["author"]

    def test_raises_unknown_license(self, lookups):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/inexistent-iri", "copy", "auth_id", None)
        msg = regex.escape(
            "The license 'http://rdfh.ch/licenses/inexistent-iri' used for an image or iiif-uri is unknown. "
            "See documentation for accepted pre-defined licenses."
        )
        with pytest.raises(InputError, match=msg):
            _get_metadata(metadata, lookups)

    def test_raises_unknown_author(self, lookups):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "unknown", "open")
        msg = regex.escape(
            "The authorship id 'unknown' referenced in a multimedia file or iiif-uri is unknown. "
            "Ensure that all referenced ids are defined in the `<authorship>` elements of your XML file."
        )
        with pytest.raises(InputError, match=msg):
            _get_metadata(metadata, lookups)


class TestTransformValues:
    def test_bool_value(self, bool_value, lookups: IntermediaryLookups):
        transformed = _transform_one_value(bool_value, lookups)
        assert isinstance(transformed, IntermediaryBoolean)
        assert transformed.value == True  # noqa:E712 (Avoid equality comparisons)
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_bool_value_with_comment(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "false", KnoraValueType.BOOLEAN_VALUE, None, "comment")
        transformed = _transform_one_value(val, lookups)
        assert transformed.value == False  # noqa:E712 (Avoid equality comparisons)
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert transformed.comment == "comment"

    def test_bool_value_with_permissions(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, "open", None)
        transformed = _transform_one_value(val, lookups)
        assert transformed.value == True  # noqa:E712 (Avoid equality comparisons)
        assert transformed.prop_iri == HAS_PROP
        assert isinstance(transformed.permissions, Permissions)
        assert not transformed.comment

    def test_bool_value_with_non_existing_permissions(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, "nonExisting", None)
        with pytest.raises(PermissionNotExistsError):
            _transform_one_value(val, lookups)

    def test_color_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "#5d1f1e", KnoraValueType.COLOR_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryColor)
        assert transformed.value == "#5d1f1e"
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_date_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "CE:1849:CE:1850", KnoraValueType.DATE_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryDate)
        assert isinstance(transformed.value, Date)
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_decimal_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "1.4", KnoraValueType.DECIMAL_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryDecimal)
        assert transformed.value == 1.4
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_geometry_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(f"{KNORA_API_STR}hasGeometry", "{}", KnoraValueType.GEOM_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryGeometry)
        assert transformed.value == "{}"
        assert transformed.prop_iri == f"{KNORA_API_STR}hasGeometry"
        assert not transformed.permissions
        assert not transformed.comment

    def test_geoname_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "5416656", KnoraValueType.GEONAME_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryGeoname)
        assert transformed.value == "5416656"
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_integer_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "1", KnoraValueType.INT_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryInt)
        assert transformed.value == 1
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_interval_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(f"{KNORA_API_STR}hasSegmentBounds", ("1", "2"), KnoraValueType.INTERVAL_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryInterval)
        assert transformed.value.start == 1.0
        assert transformed.value.end == 2.0
        assert transformed.prop_iri == f"{KNORA_API_STR}hasSegmentBounds"
        assert not transformed.permissions
        assert not transformed.comment

    def test_list_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, ("list", "node"), KnoraValueType.LIST_VALUE, "open", "cmt")
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryList)
        assert transformed.value == "http://rdfh.ch/9999/node"
        assert transformed.prop_iri == HAS_PROP
        assert isinstance(transformed.permissions, Permissions)
        assert transformed.comment == "cmt"

    def test_simple_text_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "text", KnoraValueType.SIMPLETEXT_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediarySimpleText)
        assert transformed.value == "text"
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_richtext_value(self, lookups: IntermediaryLookups):
        text_str = "<text>this is text</text>"
        val = ParsedValue(HAS_PROP, text_str, KnoraValueType.RICHTEXT_VALUE, "open", "cmt")
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryRichtext)
        assert transformed.value.xmlstr == text_str
        assert transformed.prop_iri == HAS_PROP
        assert isinstance(transformed.permissions, Permissions)
        assert transformed.comment == "cmt"
        assert transformed.resource_references == set()

    def test_richtext_value_with_standoff(self, lookups: IntermediaryLookups):
        text_str = 'Comment with <a class="salsah-link" href="IRI:link:IRI">link text</a>.'
        val = ParsedValue(HAS_PROP, text_str, KnoraValueType.RICHTEXT_VALUE, "open", "cmt")
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryRichtext)
        assert transformed.value.xmlstr == text_str
        assert transformed.prop_iri == HAS_PROP
        assert isinstance(transformed.permissions, Permissions)
        assert transformed.comment == "cmt"
        assert transformed.resource_references == {"link"}

    def test_link_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "other_id", KnoraValueType.LINK_VALUE, "open", "cmt")
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryLink)
        assert transformed.value == "other_id"
        assert transformed.prop_iri == f"{HAS_PROP}Value"
        assert isinstance(transformed.permissions, Permissions)
        assert transformed.comment == "cmt"

    def test_time_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "2019-10-23T13:45:12.01-14:00", KnoraValueType.TIME_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryTime)
        assert transformed.value == "2019-10-23T13:45:12.01-14:00"
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment

    def test_uri_value(self, lookups: IntermediaryLookups):
        val = ParsedValue(HAS_PROP, "https://dasch.swiss", KnoraValueType.URI_VALUE, None, None)
        transformed = _transform_one_value(val, lookups)
        assert isinstance(transformed, IntermediaryUri)
        assert transformed.value == "https://dasch.swiss"
        assert transformed.prop_iri == HAS_PROP
        assert not transformed.permissions
        assert not transformed.comment


class TestPermissions:
    def test_good(self, lookups):
        result = _resolve_permission("open", lookups.permissions)
        assert isinstance(result, Permissions)
        assert str(result) == "CR knora-admin:ProjectAdmin"

    def test_none(self, lookups):
        result = _resolve_permission(None, lookups.permissions)
        assert not result

    def test_empty_string(self, lookups):
        result = _resolve_permission("", lookups.permissions)
        assert not result

    def test_raises(self, lookups):
        msg = regex.escape("Could not find permissions for value: inexistent")
        with pytest.raises(PermissionNotExistsError, match=msg):
            _resolve_permission("inexistent", lookups.permissions)



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from dsp_tools.commands.excel2json.lists.make_lists import _add_nodes_to_parent
from dsp_tools.commands.excel2json.lists.make_lists import _complete_id_one_df
from dsp_tools.commands.excel2json.lists.make_lists import _construct_non_duplicate_id_string
from dsp_tools.commands.excel2json.lists.make_lists import _create_auto_id_one_df
from dsp_tools.commands.excel2json.lists.make_lists import _fill_parent_id_col_one_df
from dsp_tools.commands.excel2json.lists.make_lists import _get_lang_dict
from dsp_tools.commands.excel2json.lists.make_lists import _make_list_nodes_from_df
from dsp_tools.commands.excel2json.lists.make_lists import _make_one_list
from dsp_tools.commands.excel2json.lists.make_lists import _make_one_node
from dsp_tools.commands.excel2json.lists.make_lists import _remove_duplicate_ids_in_all_excels
from dsp_tools.commands.excel2json.lists.make_lists import _resolve_duplicate_ids_all_excels
from dsp_tools.commands.excel2json.lists.make_lists import _resolve_duplicate_ids_for_auto_id_one_df
from dsp_tools.commands.excel2json.lists.make_lists import _resolve_duplicate_ids_keep_custom_change_auto_id_one_df
from dsp_tools.commands.excel2json.lists.models.deserialise import Columns
from dsp_tools.commands.excel2json.lists.models.deserialise import ExcelSheet
from dsp_tools.commands.excel2json.lists.models.serialise import ListNode
from dsp_tools.commands.excel2json.lists.models.serialise import ListRoot


class TestDuplicateID:
    def test_resolve_duplicates_in_all_excels(self, cols_en_1: Columns) -> None:
        f1_s1 = pd.DataFrame(
            {
                "id": ["0", "1", "2"],
                "en_list": ["List1", "List1", "List1"],
                "en_1": [pd.NA, "Node1", "Node2"],
                "id (optional)": [pd.NA, pd.NA, pd.NA],
            }
        )
        f2_s2 = pd.DataFrame(
            {
                "id": ["00", "1"],
                "en_list": ["List2", "List2"],
                "en_1": [pd.NA, "Node1"],
                "id (optional)": [pd.NA, pd.NA],
            }
        )
        all_excels = [
            ExcelSheet(excel_name="file1", sheet_name="sheet1", df=f1_s1, col_info=cols_en_1),
            ExcelSheet(excel_name="file2", sheet_name="sheet2", df=f2_s2, col_info=cols_en_1),
        ]
        res = _remove_duplicate_ids_in_all_excels(["1"], all_excels)
        assert len(res) == 2
        sheet_one = res[0]
        assert sheet_one.excel_name == "file1"
        assert sheet_one.df["id"].to_list() == ["0", "List1:Node1", "2"]
        sheet_two = res[1]
        assert sheet_two.excel_name == "file2"
        assert sheet_two.df["id"].to_list() == ["00", "List2:Node1"]

    def test_resolve_duplicates_in_all_excels_custom_id(self, cols_en_1: Columns) -> None:
        f1_s1 = pd.DataFrame(
            {
                "id": ["0", "1", "2"],
                "en_list": ["List1", "List1", "List1"],
                "en_1": [pd.NA, "Node1", "Node2"],
                "id (optional)": [pd.NA, "1", pd.NA],
            }
        )
        f2_s2 = pd.DataFrame(
            {
                "id": ["00", "1"],
                "en_list": ["List2", "List2"],
                "en_1": [pd.NA, "Node1"],
                "id (optional)": [pd.NA, pd.NA],
            }
        )
        all_excels = [
            ExcelSheet(excel_name="file1", sheet_name="sheet1", df=f1_s1, col_info=cols_en_1),
            ExcelSheet(excel_name="file2", sheet_name="sheet2", df=f2_s2, col_info=cols_en_1),
        ]
        res = _remove_duplicate_ids_in_all_excels(["1"], all_excels)
        assert len(res) == 2
        sheet_one = res[0]
        assert sheet_one.excel_name == "file1"
        assert sheet_one.df["id"].to_list() == ["0", "1", "2"]
        sheet_two = res[1]
        assert sheet_two.excel_name == "file2"
        assert sheet_two.df["id"].to_list() == ["00", "List2:Node1"]

    def test_analyse_resolve_all_excel_duplicates_with_duplicates(self, cols_en_1: Columns) -> None:
        f1_s1 = pd.DataFrame(
            {
                "id": ["0", "1", "2"],
                "en_list": ["List1", "List1", "List1"],
                "en_1": [pd.NA, "Node1", "Node2"],
                "id (optional)": [pd.NA, pd.NA, pd.NA],
            }
        )
        f2_s2 = pd.DataFrame(
            {
                "id": ["00", "1"],
                "en_list": ["List2", "List2"],
                "en_1": [pd.NA, "Node1"],
                "id (optional)": [pd.NA, pd.NA],
            }
        )
        all_excels = [
            ExcelSheet(excel_name="file1", sheet_name="sheet1", df=f1_s1, col_info=cols_en_1),
            ExcelSheet(excel_name="file2", sheet_name="sheet2", df=f2_s2, col_info=cols_en_1),
        ]
        res = _resolve_duplicate_ids_all_excels(all_excels)
        assert len(res) == 2
        sheet_one = res[0]
        assert sheet_one.excel_name == "file1"
        assert sheet_one.df["id"].to_list() == ["0", "List1:Node1", "2"]
        sheet_two = res[1]
        assert sheet_two.excel_name == "file2"
        assert sheet_two.df["id"].to_list() == ["00", "List2:Node1"]

    def test_analyse_resolve_all_excel_duplicates_no_duplicates(self, cols_en_1: Columns) -> None:
        f1_s1 = pd.DataFrame(
            {
                "id": ["0", "11", "2"],
                "en_list": ["List1", "List1", "List1"],
                "en_1": [pd.NA, "Node1", "Node2"],
                "id (optional)": [pd.NA, pd.NA, pd.NA],
            }
        )
        f2_s2 = pd.DataFrame(
            {
                "id": ["00", "1"],
                "en_list": ["List2", "List2"],
                "en_1": [pd.NA, "Node1"],
                "id (optional)": [pd.NA, pd.NA],
            }
        )
        all_excels = [
            ExcelSheet(excel_name="file1", sheet_name="sheet1", df=f1_s1, col_info=cols_en_1),
            ExcelSheet(excel_name="file2", sheet_name="sheet2", df=f2_s2, col_info=cols_en_1),
        ]
        res = _resolve_duplicate_ids_all_excels(all_excels)
        assert len(res) == 2
        sheet_one = res[0]
        assert sheet_one.excel_name == "file1"
        assert sheet_one.df["id"].to_list() == ["0", "11", "2"]
        sheet_two = res[1]
        assert sheet_two.excel_name == "file2"
        assert sheet_two.df["id"].to_list() == ["00", "1"]


class TestMakeOneList:
    def test_make_lists_all_good(self, cols_en_de_1_3_no_comments: Columns) -> None:
        test_df = pd.DataFrame(
            {
                "id": ["list_id", "1", "1.1", "2", "3", "3.1", "3.2", "3.2.1", "3.2.2"],
                "parent_id": ["list_id", "list_id", "1", "list_id", "list_id", "3", "3", "3.2", "3.2"],
                "en_list": [
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                    "Listname_en",
                ],
                "de_list": [
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                    "Listname_de",
                ],
                "en_1": [
                    pd.NA,
                    "Node_en_1",
                    "Node_en_1",
                    "Node_en_2",
                    "Node_en_3",
                    "Node_en_3",
                    "Node_en_3",
                    "Node_en_3",
                    "Node_en_3",
                ],
                "de_1": [
                    pd.NA,
                    "Node_de_1",
                    "Node_de_1",
                    "Node_de_2",
                    "Node_de_3",
                    "Node_de_3",
                    "Node_de_3",
                    "Node_de_3",
                    "Node_de_3",
                ],
                "en_2": [
                    pd.NA,
                    pd.NA,
                    "Node_en_1.1",
                    pd.NA,
                    pd.NA,
                    "Node_en_3.1",
                    "Node_en_3.2",
                    "Node_en_3.2",
                    "Node_en_3.2",
                ],
                "de_2": [
                    pd.NA,
                    pd.NA,
                    "Node_de_1.1",
                    pd.NA,
                    pd.NA,
                    "Node_de_3.1",
                    "Node_de_3.2",
                    "Node_de_3.2",
                    "Node_de_3.2",
                ],
                "en_3": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "Node_en_3.2.1", "Node_en_3.2.2"],
                "de_3": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "Node_de_3.2.1", "Node_de_3.2.2"],
            }
        )
        test_sheet = ExcelSheet(
            excel_name="excel", sheet_name="Sheet1", df=test_df, col_info=cols_en_de_1_3_no_comments
        )
        res = _make_one_list(test_sheet)
        assert isinstance(res, ListRoot)
        assert res.id_ == "list_id"
        assert res.labels == {"en": "Listname_en", "de": "Listname_de"}
        assert len(res.nodes) == 3
        assert res.nodes[0].id_ == "1"
        assert res.nodes[1].id_ == "2"
        assert res.nodes[2].id_ == "3"
        assert res.nodes[2].sub_nodes[0].id_ == "3.1"
        assert res.nodes[2].sub_nodes[1].id_ == "3.2"
        assert res.nodes[2].sub_nodes[1].sub_nodes[0].id_ == "3.2.1"
        assert res.nodes[2].sub_nodes[1].sub_nodes[1].id_ == "3.2.2"


class TestFillIdColumn:
    def test_to_fill(self) -> None:
        test_df = pd.DataFrame(
            {
                "id (optional)": [pd.NA, "1", pd.NA, "3", pd.NA, pd.NA, pd.NA, pd.NA],
                "en_list": ["list_en", "list_en", "list_en", "list_en", "list_en", "list_en", "list_en", "list_en"],
                "en_1": [pd.NA, "nd_en_1", "nd_en_2", "nd_en_3", "nd_en_3", "nd_en_3", "nd_en_3", "nd_en_3"],
                "en_2": [pd.NA, pd.NA, pd.NA, pd.NA, "nd_en_3.1", "nd_en_3.2", "nd_en_3.2", "nd_en_3.2"],
                "en_3": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "nd_en_3.2.1", "nd_en_3.2.2"],
            }
        )
        expected = ["list_en", pd.NA, "nd_en_2", pd.NA, "nd_en_3.1", "nd_en_3.2", "nd_en_3.2.1", "nd_en_3.2.2"]
        res = _create_auto_id_one_df(test_df, "en")
        assert res["auto_id"].to_list() == expected

    def test_nothing_to_fill(self) -> None:
        test_df = pd.DataFrame(
            {
                "id (optional)": ["list_en", "1", "2", "3", "3.1", "3.2", "3.2.1", "3.2.2"],
            }
        )
        expected = pd.DataFrame(
            {
                "id (optional)": ["list_en", "1", "2", "3", "3.1", "3.2", "3.2.1", "3.2.2"],
                "auto_id": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
            }
        )
        res = _create_auto_id_one_df(test_df, "en")
        assert_frame_equal(res, expected)

    def test_handle_duplicates(self) -> None:
        test_df = pd.DataFrame(
            {
                "id (optional)": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "accusative"],
                "en_list": [
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                ],
                "en_1": [pd.NA, "singular", "singular", "singular", "singular", "plural", "plural", "plural", "plural"],
                "en_2": [pd.NA, pd.NA, "dative", "genitive", "accusative", pd.NA, "dative", "genitive", "accusative"],
            }
        )
        res = _create_auto_id_one_df(test_df, "en")
        expected = [
            "list_en",
            "singular",
            "list_en:singular:dative",
            "list_en:singular:genitive",
            "accusative",
            "plural",
            "list_en:plural:dative",
            "list_en:plural:genitive",
            pd.NA,
        ]
        assert res["auto_id"].to_list() == expected


def test_resolve_duplicate_when_combined_with_duplicates() -> None:
    test_df = pd.DataFrame(
        {
            "id (optional)": ["1", "2", pd.NA, pd.NA],
            "auto_id": [pd.NA, pd.NA, "1", "5"],
            "id": ["1", "2", "1", "5"],
            "en_list": ["list_en", "list_en", "list_en", "list_en"],
            "en_1": [pd.NA, "Node1", "Node2", "Node3"],
        }
    )
    res = _resolve_duplicate_ids_keep_custom_change_auto_id_one_df(test_df, "en")
    assert res["id"].to_list() == ["1", "2", "list_en:Node2", "5"]


def test_resolve_duplicate_when_combined_no_duplicates() -> None:
    test_df = pd.DataFrame(
        {
            "id (optional)": ["1", "2", pd.NA, pd.NA],
            "auto_id": [pd.NA, pd.NA, "4", "5"],
            "id": ["1", "2", "4", "5"],
            "en_list": ["list_en", "list_en", "list_en", "list_en"],
            "en_1": [pd.NA, "Node1", "Node2", "Node3"],
        }
    )
    res = _resolve_duplicate_ids_keep_custom_change_auto_id_one_df(test_df, "en")
    assert res["id"].to_list() == ["1", "2", "4", "5"]


class TestFillIDAndParentIDColumns:
    def test_handle_no_duplicates(self) -> None:
        test_df = pd.DataFrame(
            {
                "id (optional)": [pd.NA, pd.NA, "2", "3", pd.NA, "5", "6", pd.NA],
                "en_list": ["list_en", "list_en", "list_en", "list_en", "list_en", "list_en", "list_en", "list_en"],
                "en_1": ["list_en", "nd_en_1", "nd_en_2", "nd_en_3", "nd_en_4", "nd_en_5", "nd_en_6", "nd_en_7"],
            }
        )
        res = _complete_id_one_df(test_df, "en")
        assert res["id"].tolist() == ["list_en", "nd_en_1", "2", "3", "nd_en_4", "5", "6", "nd_en_7"]

    def test_handle_duplicates(self) -> None:
        test_df = pd.DataFrame(
            {
                "id (optional)": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "accusative"],
                "en_list": [
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                    "list_en",
                ],
                "en_1": [pd.NA, "singular", "singular", "singular", "singular", "plural", "plural", "plural", "plural"],
                "en_2": [pd.NA, pd.NA, "dative", "genitive", "accusative", pd.NA, "dative", "genitive", "accusative"],
            }
        )
        res = _complete_id_one_df(test_df, "en")
        expected = [
            "list_en",
            "singular",
            "list_en:singular:dative",
            "list_en:singular:genitive",
            "list_en:singular:accusative",
            "plural",
            "list_en:plural:dative",
            "list_en:plural:genitive",
            "accusative",
        ]
        assert res["id"].tolist() == expected


def test_fill_parent_id() -> None:
    test_df = pd.DataFrame(
        {
            "id": ["list_en", "1", "2", "3", "3.1", "3.2", "3.2.1", "3.2.2"],
            "en_list": ["list_en", "list_en", "list_en", "list_en", "list_en", "list_en", "list_en", "list_en"],
            "en_1": [pd.NA, "nd_en_1", "nd_en_2", "nd_en_3", "nd_en_3", "nd_en_3", "nd_en_3", "nd_en_3"],
            "en_2": [pd.NA, pd.NA, pd.NA, pd.NA, "nd_en_3.1", "nd_en_3.2", "nd_en_3.2", "nd_en_3.2"],
            "en_3": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "nd_en_3.2.1", "nd_en_3.2.2"],
        }
    )
    expected = ["list_en", "list_en", "list_en", "list_en", "3", "3", "3.2", "3.2"]
    res = _fill_parent_id_col_one_df(test_df, "en")
    assert res["parent_id"].to_list() == expected


def test_fill_parent_id_same_node_names() -> None:
    test_df = pd.DataFrame(
        {
            "id": [
                "list_en",
                "singular",
                "list_en:singular:dative",
                "list_en:singular:genitive",
                "plural",
                "list_en:plural:dative",
                "list_en:plural:genitive",
            ],
            "en_list": [
                "list_en",
                "list_en",
                "list_en",
                "list_en",
                "list_en",
                "list_en",
                "list_en",
            ],
            "en_1": [pd.NA, "singular", "singular", "singular", "plural", "plural", "plural"],
            "en_2": [pd.NA, pd.NA, "dative", "genitive", pd.NA, "dative", "genitive"],
        }
    )
    res = _fill_parent_id_col_one_df(test_df, "en")
    expected = ["list_en", "list_en", "singular", "singular", "list_en", "plural", "plural"]
    assert res["parent_id"].to_list() == expected


def test_add_nodes_to_parent() -> None:
    nd_1 = ListNode("1", {"en": "Node_en_1", "de": "Node_de_1"}, comments={}, parent_id="list_id")
    nd_2 = ListNode("2", {"en": "Node_en_2", "de": "Node_de_2"}, comments={}, parent_id="list_id")
    nd_11 = ListNode("1.1", {"en": "Node_en_1.1", "de": "Node_de_1.1"}, comments={}, parent_id="1")
    nd_12 = ListNode("1.2", {"en": "Node_en_2.1", "de": "Node_de_1.2"}, comments={}, parent_id="1")
    test_dict = {"1": nd_1, "2": nd_2, "1.1": nd_11, "1.2": nd_12}
    expected = [nd_1, nd_2]
    res = _add_nodes_to_parent(test_dict, "list_id")
    assert res == expected
    assert nd_1.sub_nodes == [nd_11, nd_12]
    assert not nd_2.sub_nodes
    assert not nd_11.sub_nodes
    assert not nd_12.sub_nodes


def test_make_list_nodes_with_valid_data(cols_en_de_1_2_no_comment: Columns) -> None:
    data = {
        "id": ["list_id", "id_1", "id_1.1", "id_2"],
        "parent_id": ["list_id", "list_id", "id_1", "list_id"],
        "en_list": ["Listname_en", "Listname_en", "Listname_en", "Listname_en"],
        "de_list": ["Listname_de", "Listname_de", "Listname_de", "Listname_de"],
        "en_1": [pd.NA, "Node1", "Node1", "Node2"],
        "de_1": [pd.NA, "Knoten1", "Knoten1", "Knoten2"],
        "en_2": [pd.NA, pd.NA, "Node1.1", pd.NA],
        "de_2": [pd.NA, pd.NA, "Knoten1.1", pd.NA],
    }
    df = pd.DataFrame(data)
    node_dict = _make_list_nodes_from_df(df, cols_en_de_1_2_no_comment)
    assert len(node_dict) == 3
    one = node_dict["id_1"]
    assert isinstance(one, ListNode)
    assert one.id_ == "id_1"
    assert one.labels == {"en": "Node1", "de": "Knoten1"}
    assert one.parent_id == "list_id"
    assert not one.sub_nodes
    one_one = node_dict["id_1.1"]
    assert isinstance(one_one, ListNode)
    assert one_one.id_ == "id_1.1"
    assert one_one.labels == {"en": "Node1.1", "de": "Knoten1.1"}
    assert one_one.parent_id == "id_1"
    assert not one_one.sub_nodes
    two = node_dict["id_2"]
    assert isinstance(two, ListNode)
    assert two.id_ == "id_2"
    assert two.labels == {"en": "Node2", "de": "Knoten2"}
    assert two.parent_id == "list_id"
    assert not two.sub_nodes


class TestMakeOneNode:
    def test_all_good_first(self, cols_en_de_1_2_no_comment: Columns) -> None:
        test_series = pd.Series(
            {
                "id": "node_id",
                "parent_id": "list_id",
                "en_list": "Listname_en",
                "de_list": "Listname_de",
                "en_1": "Node_en_1",
                "de_1": "Node_de_1",
                "en_2": pd.NA,
                "de_2": pd.NA,
            }
        )
        nd = _make_one_node(test_series, cols_en_de_1_2_no_comment)
        assert isinstance(nd, ListNode)
        assert nd.id_ == "node_id"
        assert nd.labels == {"en": "Node_en_1", "de": "Node_de_1"}
        assert not nd.sub_nodes

    def test_all_good_second(self, cols_en_de_1_2_no_comment: Columns) -> None:
        test_series = pd.Series(
            {
                "id": "node_id",
                "parent_id": "list_id",
                "en_list": "Listname_en",
                "en_1": "Node_en_1",
                "de_1": "Node_de_1",
                "en_2": "Node_en_2",
                "de_2": "Node_de_2",
            }
        )
        nd = _make_one_node(test_series, cols_en_de_1_2_no_comment)
        assert isinstance(nd, ListNode)
        assert nd.id_ == "node_id"
        assert nd.labels == {"en": "Node_en_2", "de": "Node_de_2"}
        assert not nd.sub_nodes


def test_resolve_duplicates_in_one_df() -> None:
    df = pd.DataFrame(
        {"auto_id": ["id_1", "id_1", "id_2"], "en_list": ["list", "list", "list"], "en_1": ["Node1", "Node2", "Node3"]}
    )
    res = _resolve_duplicate_ids_for_auto_id_one_df(df, "en")
    assert res["auto_id"].to_list() == ["list:Node1", "list:Node2", "id_2"]


def test_construct_non_duplicate_id() -> None:
    row = pd.Series(["en_list", "en_1", "en_2", pd.NA], index=["en_list", "en_1", "en_2", "en_3"])
    assert _construct_non_duplicate_id_string(row, "en") == "en_list:en_1:en_2"


class TestGetLabels:
    def test_correct_labels_for_all_languages(self) -> None:
        row = pd.Series(
            {
                "en_1": "Hello",
                "de_1": "Hallo",
                "fr_1": "Bonjour",
                "it_1": "Ciao",
                "rm_1": pd.NA,
                "en_2": "other",
                "de_2": pd.NA,
            }
        )
        cols = ["en_1", "de_1", "fr_1", "it_1", "rm_1"]
        expected = {"en": "Hello", "de": "Hallo", "fr": "Bonjour", "it": "Ciao"}
        assert _get_lang_dict(row, cols) == expected

    def test_returns_empty_dict_for_no_languages(self) -> None:
        row = pd.Series(
            {
                "en_1": pd.NA,
                "de_1": pd.NA,
                "fr_1": pd.NA,
                "fr_2": pd.NA,
                "it_2": pd.NA,
                "rm_2": pd.NA,
            }
        )
        cols = ["en_1", "de_1", "fr_1"]
        assert not _get_lang_dict(row, cols)

    def test_parent(self) -> None:
        row = pd.Series(
            {
                "en_list": "english",
                "de_list": "german",
                "fr_1": pd.NA,
                "fr_2": pd.NA,
                "it_2": pd.NA,
                "rm_2": pd.NA,
            }
        )
        cols = ["en_list", "de_list"]
        assert _get_lang_dict(row, cols) == {"en": "english", "de": "german"}


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from rdflib import Graph

from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.validate_data import _check_for_unknown_resource_classes
from dsp_tools.commands.validate_data.validate_data import _get_all_onto_classes
from dsp_tools.commands.validate_data.validate_data import _get_all_used_classes
from test.unittests.commands.validate_data.constants import PREFIXES


@pytest.fixture
def onto() -> Graph:
    ttl = f"""{PREFIXES}
    onto:One a owl:Class ;
            knora-api:isResourceClass true .
    
    knora-api:TextValue a owl:Class ;
            knora-api:isValueClass true .
    """
    g = Graph()
    g.parse(data=ttl, format="turtle")
    return g


@pytest.fixture
def data_ok() -> Graph:
    ttl = f"""{PREFIXES}
    <http://data/identical_text_different_prop> a onto:One ;
            onto:testSimpleText <http://data/textValue> .
    
    <http://data/textValue> a knora-api:TextValue ;
            knora-api:valueAsString "Text"^^xsd:string .
    """
    g = Graph()
    g.parse(data=ttl, format="turtle")
    return g


@pytest.fixture
def data_wrong() -> Graph:
    ttl = f"""{PREFIXES}
    <http://data/identical_text_different_prop> a onto:NonExistent ;
            onto:testSimpleText <http://data/textValue> .

    <http://data/textValue> a knora-api:TextValue ;
            knora-api:valueAsString "Text"^^xsd:string .
    """
    g = Graph()
    g.parse(data=ttl, format="turtle")
    return g


@pytest.fixture
def data_prefix_non_existent() -> Graph:
    ttl = f"""{PREFIXES}
    @prefix  non-existing-onto: <http://0.0.0.0:3333/ontology/9999/non-existent/v2#> .
    
    <http://data/identical_text_different_prop> a non-existing-onto:One ;
            onto:testSimpleText <http://data/textValue> .
    
    <http://data/textValue> a knora-api:TextValue ;
            knora-api:valueAsString "Text"^^xsd:string .
    """
    g = Graph()
    g.parse(data=ttl, format="turtle")
    return g


def _get_rdf_graphs(data_graph: Graph) -> RDFGraphs:
    ttl = f"""{PREFIXES}
    onto:One a owl:Class ;
            knora-api:isResourceClass true .

    knora-api:TextValue a owl:Class ;
            knora-api:isValueClass true .
    """
    onto_g = Graph()
    onto_g.parse(data=ttl, format="turtle")
    return RDFGraphs(
        data=data_graph, ontos=onto_g, cardinality_shapes=Graph(), content_shapes=Graph(), knora_api=Graph()
    )


class TestFindUnknownClasses:
    def test_check_for_unknown_resource_classes_data_ok(self, data_ok: Graph) -> None:
        graphs = _get_rdf_graphs(data_ok)
        result = _check_for_unknown_resource_classes(graphs)
        assert not result

    def test_check_for_unknown_resource_classes_data_wrong(self, data_wrong: Graph) -> None:
        graphs = _get_rdf_graphs(data_wrong)
        result = _check_for_unknown_resource_classes(graphs)
        assert isinstance(result, UnknownClassesInData)
        assert result.unknown_classes == {"onto:NonExistent"}
        assert result.classes_onto == {"onto:One"}
        assert result._get_unknown_ontos_msg() == ""

    def test_check_for_unknown_resource_classes_data_prefix_non_existent(self, data_prefix_non_existent: Graph) -> None:
        graphs = _get_rdf_graphs(data_prefix_non_existent)
        result = _check_for_unknown_resource_classes(graphs)
        assert isinstance(result, UnknownClassesInData)
        assert result.unknown_classes == {"non-existent:One"}
        assert result.classes_onto == {"onto:One"}
        assert result._get_unknown_ontos_msg() != ""

    def test_get_all_used_classes(self, data_ok: Graph) -> None:
        result = _get_all_used_classes(data_ok)
        expected = {"TextValue", "onto:One"}
        assert result == expected

    def test_get_all_onto_classes(self, onto: Graph) -> None:
        res_cls, value_cls = _get_all_onto_classes(onto)
        assert res_cls == {"onto:One"}
        assert value_cls == {"TextValue"}


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import warnings
from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from pathlib import Path

import regex
from rdflib import Graph

from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.error.custom_warnings import DspToolsUserWarning

LIST_SEPARATOR = "\n    - "
GRAND_SEPARATOR = "\n\n----------------------------\n"


@dataclass
class OntologyValidationProblem:
    problems: list[OntologyResourceProblem]

    def get_msg(self) -> str:
        probs = sorted(self.problems, key=lambda x: x.res_iri)
        problems = [x.get_msg() for x in probs]
        return (
            "The ontology structure contains errors that prevent the validation of the data.\n"
            "Please correct the following errors and re-upload the corrected ontology.\n"
            f"Once those two steps are done, the command `validate-data` will find any problems in the data.\n"
            f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(problems)}"
        )


@dataclass
class OntologyResourceProblem:
    res_iri: str
    msg: str

    def get_msg(self) -> str:
        return f"Resource Class: {self.res_iri} | Problem: {self.msg}"


@dataclass
class UnknownClassesInData:
    unknown_classes: set[str]
    classes_onto: set[str]

    def get_msg(self) -> str:
        if unknown := self._get_unknown_ontos_msg():
            return unknown
        return self._get_unknown_classes_msg()

    def _get_unknown_ontos_msg(self) -> str:
        def split_prefix(relative_iri: str) -> str:
            return relative_iri.split(":")[0]

        used_ontos = set(split_prefix(x) for x in self.unknown_classes)
        exising_ontos = set(split_prefix(x) for x in self.classes_onto)
        msg = ""
        if unknown := used_ontos - exising_ontos:
            msg = (
                f"Your data uses ontologies that don't exist in the database.\n"
                f"The following ontologies that are used in the data are unknown: {', '.join(exising_ontos)}"
                f"The following ontologies are uploaded: {', '.join(unknown)}\n"
            )
        return msg

    def _get_unknown_classes_msg(self) -> str:
        unknown_classes = sorted(list(self.unknown_classes))
        known_classes = sorted(list(self.classes_onto))
        return (
            f"Your data uses resource classes that do not exist in the ontologies in the database.\n"
            f"The following classes that are used in the data are unknown: {', '.join(unknown_classes)}\n"
            f"The following classes exist in the uploaded ontologies: {', '.join(known_classes)}\n"
        )


@dataclass
class UnexpectedResults:
    components: list[UnexpectedComponent]

    def save_inform_user(self, results_graph: Graph, shacl: Graph, data: Graph) -> None:
        cwdr = Path.cwd()
        prefix = f"{datetime.now()!s}_"
        unique_components = list(set(x.component_type for x in self.components))
        components = sorted(unique_components)
        msg = (
            f"Unexpected violations were found in the validation results:"
            f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(components)}\n"
            f"Please contact the development team with the files starting with the timestamp '{prefix}' "
            f"in the directory '{cwdr}'."
        )
        save_path = cwdr / f"{prefix}validation_result.ttl"
        results_graph.serialize(save_path)
        shacl_p = cwdr / f"{prefix}shacl.ttl"
        shacl.serialize(shacl_p)
        data_p = cwdr / f"{prefix}data.ttl"
        data.serialize(data_p)
        warnings.warn(DspToolsUserWarning(msg))


@dataclass
class AllProblems:
    problems: list[InputProblem]
    unexpected_results: UnexpectedResults | None


@dataclass
class InputProblem:
    problem_type: ProblemType
    res_id: str
    res_type: str
    prop_name: str
    message: str | None = None
    input_value: str | None = None
    input_type: str | None = None
    expected: str | None = None

    def __post_init__(self) -> None:
        self.message = self._clean_input(self.message)
        self.expected = self._clean_input(self.expected)

    def _clean_input(self, to_clean: str | None) -> str | None:
        if not to_clean:
            return None
        if not regex.search(r"\S+", to_clean):
            return None
        str_split = [content for x in to_clean.split(" ") if (content := x.strip())]
        return " ".join(str_split)


class ProblemType(StrEnum):
    GENERIC = "generic"
    FILE_VALUE = "file problem"
    MAX_CARD = "Maximum Cardinality Violation"
    MIN_CARD = "Minimum Cardinality Violation"
    NON_EXISTING_CARD = "The resource class does not have a cardinality for this property."
    FILE_VALUE_PROHIBITED = "A file was added to the resource. This resource type must not have a file."
    VALUE_TYPE_MISMATCH = "Value Type Mismatch"
    INPUT_REGEX = "Wrong Format of Input"
    LINK_TARGET_TYPE_MISMATCH = "Linked Resource Type Mismatch"
    INEXISTENT_LINKED_RESOURCE = "Linked Resource does not exist"
    DUPLICATE_VALUE = "Your input is duplicated"



from __future__ import annotations

import warnings
from typing import cast

from lxml import etree

from dsp_tools.error.custom_warnings import DspToolsUserWarning
from dsp_tools.error.exceptions import InputError
from dsp_tools.xmllib.constants import DASCH_SCHEMA
from dsp_tools.xmllib.constants import XML_NAMESPACE_MAP
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.dsp_base_resources import AudioSegmentResource
from dsp_tools.xmllib.models.dsp_base_resources import LinkResource
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.dsp_base_resources import VideoSegmentResource
from dsp_tools.xmllib.models.file_values import AuthorshipLookup
from dsp_tools.xmllib.models.res import Resource
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import Value
from dsp_tools.xmllib.serialise.serialise_file_value import serialise_file_value
from dsp_tools.xmllib.serialise.serialise_values import serialise_values
from dsp_tools.xmllib.type_aliases import AnyResource
from dsp_tools.xmllib.value_checkers import is_string_like


def serialise_resources(resources: list[AnyResource], authorship_lookup: AuthorshipLookup) -> list[etree._Element]:
    """
    Serialise all the resources

    Args:
        resources: list of resources
        authorship_lookup: lookup to map the authors to the corresponding IDs

    Returns:
        serialised resources
    """
    return [_serialise_one_resource(x, authorship_lookup) for x in resources]


def _serialise_one_resource(res: AnyResource, authorship_lookup: AuthorshipLookup) -> etree._Element:
    match res:
        case Resource():
            return _serialise_generic_resource(res, authorship_lookup)
        case RegionResource():
            return _serialise_region(res)
        case LinkResource():
            return _serialise_link(res)
        case AudioSegmentResource():
            return _serialise_segment(res, "audio-segment")
        case VideoSegmentResource():
            return _serialise_segment(res, "video-segment")
        case _:
            raise InputError(
                f"An unknown resource was added to the root. "
                f"Only Resource, RegionResource, LinkResource, VideoSegmentResource, AudioSegmentResource "
                f"are permitted."
                f"The input type is {res.__class__.__name__}"
            )


def _serialise_generic_resource(res: Resource, authorship_lookup: AuthorshipLookup) -> etree._Element:
    ele = _make_generic_resource_element(res, "resource")
    ele.attrib["restype"] = res.restype
    if res.file_value:
        auth_id = authorship_lookup.get_id(res.file_value.metadata.authorship)
        ele.append(serialise_file_value(res.file_value, auth_id))
    ele.extend(serialise_values(res.values))
    return ele


def _serialise_region(res: RegionResource) -> etree._Element:
    ele = _make_generic_resource_element(res, "region")
    ele.extend(_serialise_geometry_shape(res))
    props: list[Value] = [res.region_of]
    if res.comments:
        cmnt = cast(list[Value], res.comments)
        props.extend(cmnt)
    ele.extend(serialise_values(props))
    return ele


def _serialise_geometry_shape(res: RegionResource) -> list[etree._Element]:
    prop_list: list[etree._Element] = []
    if not res.geometry:
        msg = (
            f"The region resource with the ID '{res.res_id}' does not have a geometry, "
            f"please note that an xmlupload will fail."
        )
        warnings.warn(DspToolsUserWarning(msg))

        return prop_list
    geo_prop = etree.Element(f"{DASCH_SCHEMA}geometry-prop", name="hasGeometry", nsmap=XML_NAMESPACE_MAP)
    ele = etree.Element(f"{DASCH_SCHEMA}geometry", nsmap=XML_NAMESPACE_MAP)
    ele.text = res.geometry.to_json_string()
    geo_prop.append(ele)
    prop_list.append(geo_prop)
    prop_list.extend(
        serialise_values([ColorValue(value=res.geometry.color, prop_name="hasColor", resource_id=res.res_id)]),
    )
    return prop_list


def _serialise_link(res: LinkResource) -> etree._Element:
    problems = []
    if not res.comments:
        problems.append("at least one comment")
    if not res.link_to:
        problems.append("at least two links")
    if problems:
        msg = (
            f"The link object with the ID '{res.res_id}' requires {' and '.join(problems)}. "
            f"Please note that an xmlupload will fail."
        )
        warnings.warn(DspToolsUserWarning(msg))
    ele = _make_generic_resource_element(res, "link")
    comments = cast(list[Value], res.comments)
    links_to = cast(list[Value], res.link_to)
    generic_vals = comments + links_to
    ele.extend(serialise_values(generic_vals))
    return ele


def _serialise_segment(res: AudioSegmentResource | VideoSegmentResource, segment_type: str) -> etree._Element:
    _validate_segment(res)
    seg = _make_generic_resource_element(res, segment_type)
    seg.extend(_serialise_segment_children(res))
    return seg


def _make_generic_resource_element(res: AnyResource, res_type: str) -> etree._Element:
    attribs = {"label": res.label, "id": res.res_id}
    if res.permissions != Permissions.PROJECT_SPECIFIC_PERMISSIONS:
        attribs["permissions"] = res.permissions.value
    return etree.Element(f"{DASCH_SCHEMA}{res_type}", attrib=attribs, nsmap=XML_NAMESPACE_MAP)


def _validate_segment(segment: AudioSegmentResource | VideoSegmentResource) -> None:
    problems = []
    if not is_string_like(segment.res_id):
        problems.append(f"Field: Resource ID | Value: {segment.res_id}")
    if not is_string_like(segment.label):
        problems.append(f"Field: label | Value: {segment.label}")
    if not is_string_like(segment.segment_of):
        problems.append(f"Field: segment_of | Value: {segment.segment_of}")
    if segment.title and not is_string_like(segment.title):
        problems.append(f"Field: title | Value: {segment.title}")
    if fails := [x for x in segment.comments if not is_string_like(x)]:
        problems.extend([f"Field: comment | Value: {x}" for x in fails])
    if fails := [x for x in segment.descriptions if not is_string_like(x)]:
        problems.extend([f"Field: description | Value: {x}" for x in fails])
    if fails := [x for x in segment.keywords if not is_string_like(x)]:
        problems.extend([f"Field: keywords | Value: {x}" for x in fails])
    if fails := [x for x in segment.relates_to if not is_string_like(x)]:
        problems.extend([f"Field: relates_to | Value: {x}" for x in fails])
    if problems:
        msg = f"The resource with the ID '{segment.res_id}' has the following problem(s):{'\n- '.join(problems)}"
        warnings.warn(DspToolsUserWarning(msg))


def _serialise_segment_children(segment: AudioSegmentResource | VideoSegmentResource) -> list[etree._Element]:
    segment_elements = []
    segment_of = etree.Element(f"{DASCH_SCHEMA}isSegmentOf", nsmap=XML_NAMESPACE_MAP)
    segment_of.text = segment.segment_of
    segment_elements.append(segment_of)
    segment_elements.append(
        etree.Element(
            f"{DASCH_SCHEMA}hasSegmentBounds",
            attrib={
                "segment_start": str(segment.segment_bounds.segment_start),
                "segment_end": str(segment.segment_bounds.segment_end),
            },
            nsmap=XML_NAMESPACE_MAP,
        )
    )
    if segment.title:
        segment_elements.append(_make_element_with_text("hasTitle", segment.title))
    segment_elements.extend([_make_element_with_text("hasComment", x) for x in segment.comments])
    segment_elements.extend([_make_element_with_text("hasDescription", x) for x in segment.descriptions])
    segment_elements.extend([_make_element_with_text("hasKeyword", x) for x in segment.keywords])
    segment_elements.extend([_make_element_with_text("relatesTo", x) for x in segment.relates_to])
    return segment_elements


def _make_element_with_text(tag_name: str, text_content: str) -> etree._Element:
    ele = etree.Element(f"{DASCH_SCHEMA}{tag_name}", nsmap=XML_NAMESPACE_MAP)
    ele.text = text_content
    return ele



from dataclasses import dataclass

from dsp_tools.error.problems import Problem


@dataclass
class IllegalTagProblem(Problem):
    orig_err_msg: str
    pseudo_xml: str

    def execute_error_protocol(self) -> str:
        msg = (
            "The XML tags contained in a richtext property (encoding=xml) must be well-formed. "
            "The special characters <, > and & are only allowed to construct a tag. "
        )
        msg += f"\nOriginal error message: {self.orig_err_msg}"
        msg += f"\nEventual line/column numbers are relative to this text: {self.pseudo_xml}"
        return msg



import json
import urllib.parse
from pathlib import Path
from typing import Any
from typing import Iterator

import pytest
import requests
from rdflib import RDF
from rdflib import RDFS
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.xmlupload.xmlupload import xmlupload
from dsp_tools.utils.rdflib_constants import KNORA_API
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts
from test.e2e.setup_testcontainers.setup import get_containers

PROJECT_SHORTCODE = "4125"
ONTO_NAME = "e2e-testonto"
PROPS_IN_ONTO_JSON = 1
RESCLASSES_IN_ONTO_JSON = 2


@pytest.fixture(scope="module")
def container_ports() -> Iterator[ExternalContainerPorts]:
    with get_containers() as metadata:
        yield metadata.ports


@pytest.fixture(scope="module")
def creds(container_ports: ExternalContainerPorts) -> ServerCredentials:
    return ServerCredentials(
        "root@example.com",
        "test",
        f"http://0.0.0.0:{container_ports.api}",
        f"http://0.0.0.0:{container_ports.ingest}",
    )


@pytest.fixture(scope="module")
def onto_iri(creds: ServerCredentials) -> str:
    return f"{creds.server}/ontology/{PROJECT_SHORTCODE}/{ONTO_NAME}/v2"


@pytest.fixture(scope="module")
def _create_project(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/json-project/test-project-e2e.json"), creds, verbose=True)


@pytest.fixture(scope="module")
def _xmlupload(_create_project: None, creds: ServerCredentials) -> Iterator[None]:
    assert xmlupload(Path("testdata/xml-data/test-data-e2e.xml"), creds, ".")
    yield
    if found := list(Path.cwd().glob(f"id2iri_{PROJECT_SHORTCODE}_localhost_*.json")):
        found[0].unlink()


@pytest.fixture(scope="module")
def auth_header(_create_project: None, creds: ServerCredentials) -> dict[str, str]:
    payload = {"email": creds.user, "password": creds.password}
    token: str = requests.post(f"{creds.server}/v2/authentication", json=payload, timeout=3).json()["token"]
    return {"Authorization": f"Bearer {token}"}


@pytest.fixture(scope="module")
def project_iri(_create_project: None, creds: ServerCredentials) -> str:
    get_project_route = f"{creds.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project_iri: str = requests.get(get_project_route, timeout=3).json()["project"]["id"]
    return project_iri


@pytest.mark.usefixtures("_create_project")
def test_project(auth_header: dict[str, str], creds: ServerCredentials, onto_iri: str) -> None:
    get_project_url = f"{creds.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project = requests.get(get_project_url, headers=auth_header, timeout=3).json()["project"]
    _check_project(project, onto_iri)

    onto = requests.get(onto_iri, headers=auth_header, timeout=3).json()["@graph"]
    _check_props([elem for elem in onto if elem.get("knora-api:isResourceProperty")])
    _check_resclasses([elem for elem in onto if elem.get("knora-api:isResourceClass")])


@pytest.mark.usefixtures("_xmlupload")
def test_xmlupload(auth_header: dict[str, str], project_iri: str, creds: ServerCredentials, onto_iri: str) -> None:
    img_resources = _get_resources(f"{onto_iri}#ImageResource", auth_header, project_iri, creds)
    _analyze_img_resources(img_resources)
    pdf_resources = _get_resources(f"{onto_iri}#PDFResource", auth_header, project_iri, creds)
    _analyze_pdf_resources(pdf_resources)


@pytest.mark.usefixtures("_xmlupload")
def test_all_copyright_holders(auth_header: dict[str, str], creds: ServerCredentials) -> None:
    response = _get_copyright_holders(auth_header, creds)
    assert set(response["data"]) == {"DaSCH", "Wellcome Collection"}


def _get_copyright_holders(auth_header: dict[str, str], creds: ServerCredentials) -> dict[str, Any]:
    url = f"{creds.server}/admin/projects/shortcode/4125/legal-info/copyright-holders?page=1&page-size=25&order=Asc"
    headers = auth_header | {"Accept": "application/json"}
    response = requests.get(url=url, headers=headers, timeout=3).json()
    return dict(response)


def _check_project(project: dict[str, Any], onto_iri: str) -> None:
    assert project["shortname"] == "e2e-tp"
    assert project["shortcode"] == PROJECT_SHORTCODE
    assert project["longname"] == "e2e test project"
    assert project["description"] == [{"value": "The e2e test project", "language": "en"}]
    assert project["keywords"] == ["e2e-test-projec-keyword"]
    assert project["ontologies"] == [onto_iri]


def _check_props(props: list[dict[str, Any]]) -> None:
    assert len(props) == PROPS_IN_ONTO_JSON
    assert props[0]["@id"] == f"{ONTO_NAME}:hasText"
    assert props[0]["rdfs:subPropertyOf"] == {"@id": "knora-api:hasValue"}
    assert props[0]["knora-api:objectType"] == {"@id": "knora-api:TextValue"}
    assert props[0]["rdfs:label"] == "Text"


def _check_resclasses(resclasses: list[dict[str, Any]]) -> None:
    assert len(resclasses) == RESCLASSES_IN_ONTO_JSON
    res_1, res_2 = resclasses

    assert res_1["@id"] == f"{ONTO_NAME}:ImageResource"
    assert res_1["rdfs:label"] == "Image Resource"
    cards_1 = res_1["rdfs:subClassOf"]
    fileval_card = [x for x in cards_1 if x.get("owl:onProperty", {}).get("@id") == "knora-api:hasStillImageFileValue"]
    assert len(fileval_card) == 1
    assert fileval_card[0].get("owl:cardinality") == 1
    hasText_card = [x for x in cards_1 if x.get("owl:onProperty", {}).get("@id") == f"{ONTO_NAME}:hasText"]
    assert len(hasText_card) == 1

    assert res_2["@id"] == f"{ONTO_NAME}:PDFResource"
    assert res_2["rdfs:label"] == "PDF Resource"


def _get_resources(resclass_iri: str, auth_header: dict[str, str], project_iri: str, creds: ServerCredentials) -> str:
    resclass_iri_encoded = urllib.parse.quote_plus(resclass_iri)
    get_resources_route = f"{creds.server}/v2/resources?resourceClass={resclass_iri_encoded}&page=0"
    headers = auth_header | {"X-Knora-Accept-Project": project_iri}
    response = requests.get(get_resources_route, timeout=3, headers=headers).json()
    resources = json.dumps(response)
    return resources


def _analyze_img_resources(img_resources: str) -> None:
    g = Graph()
    g.parse(data=img_resources, format="json-ld")
    open_permissions = Literal(
        "CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember|V knora-admin:KnownUser,knora-admin:UnknownUser"
    )

    labels = {str(x) for x in g.objects(predicate=RDFS.label)}
    assert labels == {"Resource 1", "Resource 2"}

    res_1_iri = next(g.subjects(RDFS.label, Literal("Resource 1")))
    assert next(g.objects(res_1_iri, KNORA_API.hasPermissions)) == open_permissions
    file_1_iri = next(g.objects(res_1_iri, KNORA_API.hasStillImageFileValue))
    assert next(g.objects(file_1_iri, RDF.type)) == KNORA_API.StillImageFileValue
    assert next(g.objects(file_1_iri, KNORA_API.hasAuthorship)) == Literal("Johannes Nussbaum")
    assert next(g.objects(file_1_iri, KNORA_API.hasCopyrightHolder)) == Literal("DaSCH")
    assert next(g.objects(file_1_iri, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-4.0")

    res_2_iri = next(g.subjects(RDFS.label, Literal("Resource 2")))
    file_2_iri = next(g.objects(res_2_iri, KNORA_API.hasStillImageFileValue))
    assert next(g.objects(file_2_iri, RDF.type)) == KNORA_API.StillImageExternalFileValue
    assert next(g.objects(file_2_iri, KNORA_API.hasPermissions)) == open_permissions
    assert next(g.objects(file_2_iri, KNORA_API.hasAuthorship)) == Literal("Cavanagh, Annie")
    assert next(g.objects(file_2_iri, KNORA_API.hasCopyrightHolder)) == Literal("Wellcome Collection")
    assert next(g.objects(file_2_iri, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-nc-4.0")
    val_with_comment = next(g.subjects(KNORA_API.valueAsString, Literal("first text value")))
    assert next(g.objects(val_with_comment, KNORA_API.valueHasComment)) == Literal("Comment")
    val_with_perm = next(g.subjects(KNORA_API.valueAsString, Literal("second text value")))
    assert next(g.objects(val_with_perm, KNORA_API.hasPermissions)) == open_permissions


def _analyze_pdf_resources(pdf_resources: str) -> None:
    g = Graph()
    g.parse(data=pdf_resources, format="json-ld")
    res_iri = next(g.subjects(RDFS.label, Literal("Resource 3")))
    file_iri = next(g.objects(res_iri, KNORA_API.hasDocumentFileValue))
    open_permissions = Literal(
        "CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember|V knora-admin:KnownUser,knora-admin:UnknownUser"
    )
    assert next(g.objects(file_iri, RDF.type)) == KNORA_API.DocumentFileValue
    assert next(g.objects(file_iri, KNORA_API.hasAuthorship)) == Literal("Nora Ammann")
    assert next(g.objects(file_iri, KNORA_API.hasCopyrightHolder)) == Literal("DaSCH")
    assert next(g.objects(file_iri, KNORA_API.hasLicense)) == URIRef("http://rdfh.ch/licenses/cc-by-4.0")
    assert next(g.objects(file_iri, KNORA_API.hasPermissions)) == open_permissions
